"""Support for dynamic COM client support.

Introduction
 Dynamic COM client support is the ability to use a COM server without
 prior knowledge of the server.  This can be used to talk to almost all
 COM servers, including much of MS Office.

 In general, you should not use this module directly - see below.

Example
 >>> import win32com.client
 >>> xl = win32com.client.Dispatch("Excel.Application")
 # The line above invokes the functionality of this class.
 # xl is now an object we can use to talk to Excel.
 >>> xl.Visible = 1 # The Excel window becomes visible.

"""
import sys
import traceback
import types

import pythoncom
import winerror
from . import build

from pywintypes import IIDType

import win32com.client  # Needed as code we eval() references it.

debugging = 0  # General debugging
debugging_attr = 0  # Debugging dynamic attribute lookups.

LCID = 0x0

# These errors generally mean the property or method exists,
# but can't be used in this context - eg, property instead of a method, etc.
# Used to determine if we have a real error or not.
ERRORS_BAD_CONTEXT = [
    winerror.DISP_E_MEMBERNOTFOUND,
    winerror.DISP_E_BADPARAMCOUNT,
    winerror.DISP_E_PARAMNOTOPTIONAL,
    winerror.DISP_E_TYPEMISMATCH,
    winerror.E_INVALIDARG,
]

ALL_INVOKE_TYPES = [
    pythoncom.INVOKE_PROPERTYGET,
    pythoncom.INVOKE_PROPERTYPUT,
    pythoncom.INVOKE_PROPERTYPUTREF,
    pythoncom.INVOKE_FUNC,
]


def debug_print(*args):
    if debugging:
        for arg in args:
            print(arg, end=" ")
        print()


def debug_attr_print(*args):
    if debugging_attr:
        for arg in args:
            print(arg, end=" ")
        print()


def MakeMethod(func, inst, cls):
    return types.MethodType(func, inst)


# get the type objects for IDispatch and IUnknown
PyIDispatchType = pythoncom.TypeIIDs[pythoncom.IID_IDispatch]
PyIUnknownType = pythoncom.TypeIIDs[pythoncom.IID_IUnknown]

_GoodDispatchTypes = (str, IIDType)
_defaultDispatchItem = build.DispatchItem


def _GetGoodDispatch(IDispatch, clsctx=pythoncom.CLSCTX_SERVER):
    # quick return for most common case
    if isinstance(IDispatch, PyIDispatchType):
        return IDispatch
    if isinstance(IDispatch, _GoodDispatchTypes):
        try:
            IDispatch = pythoncom.connect(IDispatch)
        except pythoncom.ole_error:
            IDispatch = pythoncom.CoCreateInstance(
                IDispatch, None, clsctx, pythoncom.IID_IDispatch
            )
    else:
        # may already be a wrapped class.
        IDispatch = getattr(IDispatch, "_oleobj_", IDispatch)
    return IDispatch


def _GetGoodDispatchAndUserName(IDispatch, userName, clsctx):
    # Get a dispatch object, and a 'user name' (ie, the name as
    # displayed to the user in repr() etc.
    if userName is None:
        if isinstance(IDispatch, str):
            userName = IDispatch
        ## ??? else userName remains None ???
    else:
        userName = str(userName)
    return (_GetGoodDispatch(IDispatch, clsctx), userName)


def _GetDescInvokeType(entry, invoke_type):
    # determine the wFlags argument passed as input to IDispatch::Invoke
    # Only ever called by __getattr__ and __setattr__ from dynamic objects!
    # * `entry` is a MapEntry with whatever typeinfo we have about the property we are getting/setting.
    # * `invoke_type` is either INVOKE_PROPERTYGET | INVOKE_PROPERTYSET and really just
    #   means "called by __getattr__" or "called by __setattr__"
    if not entry or not entry.desc:
        return invoke_type

    if entry.desc.desckind == pythoncom.DESCKIND_VARDESC:
        return invoke_type

    # So it's a FUNCDESC - just use what it specifies.
    return entry.desc.invkind


def Dispatch(
    IDispatch,
    userName=None,
    createClass=None,
    typeinfo=None,
    UnicodeToString=None,
    clsctx=pythoncom.CLSCTX_SERVER,
):
    assert UnicodeToString is None, "this is deprecated and will go away"
    IDispatch, userName = _GetGoodDispatchAndUserName(IDispatch, userName, clsctx)
    if createClass is None:
        createClass = CDispatch
    lazydata = None
    try:
        if typeinfo is None:
            typeinfo = IDispatch.GetTypeInfo()
        if typeinfo is not None:
            try:
                # try for a typecomp
                typecomp = typeinfo.GetTypeComp()
                lazydata = typeinfo, typecomp
            except pythoncom.com_error:
                pass
    except pythoncom.com_error:
        typeinfo = None
    olerepr = MakeOleRepr(IDispatch, typeinfo, lazydata)
    return createClass(IDispatch, olerepr, userName, lazydata=lazydata)


def MakeOleRepr(IDispatch, typeinfo, typecomp):
    olerepr = None
    if typeinfo is not None:
        try:
            attr = typeinfo.GetTypeAttr()
            # If the type info is a special DUAL interface, magically turn it into
            # a DISPATCH typeinfo.
            if (
                attr[5] == pythoncom.TKIND_INTERFACE
                and attr[11] & pythoncom.TYPEFLAG_FDUAL
            ):
                # Get corresponding Disp interface;
                # -1 is a special value which does this for us.
                href = typeinfo.GetRefTypeOfImplType(-1)
                typeinfo = typeinfo.GetRefTypeInfo(href)
                attr = typeinfo.GetTypeAttr()
            if typecomp is None:
                olerepr = build.DispatchItem(typeinfo, attr, None, 0)
            else:
                olerepr = build.LazyDispatchItem(attr, None)
        except pythoncom.ole_error:
            pass
    if olerepr is None:
        olerepr = build.DispatchItem()
    return olerepr


def DumbDispatch(
    IDispatch,
    userName=None,
    createClass=None,
    UnicodeToString=None,
    clsctx=pythoncom.CLSCTX_SERVER,
):
    "Dispatch with no type info"
    assert UnicodeToString is None, "this is deprecated and will go away"
    IDispatch, userName = _GetGoodDispatchAndUserName(IDispatch, userName, clsctx)
    if createClass is None:
        createClass = CDispatch
    return createClass(IDispatch, build.DispatchItem(), userName)


class CDispatch:
    def __init__(
        self, IDispatch, olerepr, userName=None, UnicodeToString=None, lazydata=None
    ):
        assert UnicodeToString is None, "this is deprecated and will go away"
        if userName is None:
            userName = "<unknown>"
        self.__dict__["_oleobj_"] = IDispatch
        self.__dict__["_username_"] = userName
        self.__dict__["_olerepr_"] = olerepr
        self.__dict__["_mapCachedItems_"] = {}
        self.__dict__["_builtMethods_"] = {}
        self.__dict__["_enum_"] = None
        self.__dict__["_unicode_to_string_"] = None
        self.__dict__["_lazydata_"] = lazydata

    def __call__(self, *args):
        "Provide 'default dispatch' COM functionality - allow instance to be called"
        if self._olerepr_.defaultDispatchName:
            invkind, dispid = self._find_dispatch_type_(
                self._olerepr_.defaultDispatchName
            )
        else:
            invkind, dispid = (
                pythoncom.DISPATCH_METHOD | pythoncom.DISPATCH_PROPERTYGET,
                pythoncom.DISPID_VALUE,
            )
        if invkind is not None:
            allArgs = (dispid, LCID, invkind, 1) + args
            return self._get_good_object_(
                self._oleobj_.Invoke(*allArgs), self._olerepr_.defaultDispatchName, None
            )
        raise TypeError("This dispatch object does not define a default method")

    def __bool__(self):
        return True  # ie "if object:" should always be "true" - without this, __len__ is tried.
        # _Possibly_ want to defer to __len__ if available, but Im not sure this is
        # desirable???

    def __repr__(self):
        return "<COMObject %s>" % (self._username_)

    def __str__(self):
        # __str__ is used when the user does "print object", so we gracefully
        # fall back to the __repr__ if the object has no default method.
        try:
            return str(self.__call__())
        except pythoncom.com_error as details:
            if details.hresult not in ERRORS_BAD_CONTEXT:
                raise
            return self.__repr__()

    def __dir__(self):
        lst = list(self.__dict__.keys()) + dir(self.__class__) + self._dir_ole_()
        try:
            lst += [p.Name for p in self.Properties_]
        except AttributeError:
            pass
        return list(set(lst))

    def _dir_ole_(self):
        items_dict = {}
        for iTI in range(0, self._oleobj_.GetTypeInfoCount()):
            typeInfo = self._oleobj_.GetTypeInfo(iTI)
            self._UpdateWithITypeInfo_(items_dict, typeInfo)
        return list(items_dict.keys())

    def _UpdateWithITypeInfo_(self, items_dict, typeInfo):
        typeInfos = [typeInfo]
        # suppress IDispatch and IUnknown methods
        inspectedIIDs = {pythoncom.IID_IDispatch: None}

        while len(typeInfos) > 0:
            typeInfo = typeInfos.pop()
            typeAttr = typeInfo.GetTypeAttr()

            if typeAttr.iid not in inspectedIIDs:
                inspectedIIDs[typeAttr.iid] = None
                for iFun in range(0, typeAttr.cFuncs):
                    funDesc = typeInfo.GetFuncDesc(iFun)
                    funName = typeInfo.GetNames(funDesc.memid)[0]
                    if funName not in items_dict:
                        items_dict[funName] = None

                # Inspect the type info of all implemented types
                # E.g. IShellDispatch5 implements IShellDispatch4 which implements IShellDispatch3 ...
                for iImplType in range(0, typeAttr.cImplTypes):
                    iRefType = typeInfo.GetRefTypeOfImplType(iImplType)
                    refTypeInfo = typeInfo.GetRefTypeInfo(iRefType)
                    typeInfos.append(refTypeInfo)

    # Delegate comparison to the oleobjs, as they know how to do identity.
    def __eq__(self, other):
        other = getattr(other, "_oleobj_", other)
        return self._oleobj_ == other

    def __ne__(self, other):
        other = getattr(other, "_oleobj_", other)
        return self._oleobj_ != other

    def __int__(self):
        return int(self.__call__())

    def __len__(self):
        invkind, dispid = self._find_dispatch_type_("Count")
        if invkind:
            return self._oleobj_.Invoke(dispid, LCID, invkind, 1)
        raise TypeError("This dispatch object does not define a Count method")

    def _NewEnum(self):
        try:
            invkind = pythoncom.DISPATCH_METHOD | pythoncom.DISPATCH_PROPERTYGET
            enum = self._oleobj_.InvokeTypes(
                pythoncom.DISPID_NEWENUM, LCID, invkind, (13, 10), ()
            )
        except pythoncom.com_error:
            return None  # no enumerator for this object.
        from . import util

        return util.WrapEnum(enum, None)

    def __getitem__(self, index):  # syver modified
        # Improved __getitem__ courtesy Syver Enstad
        # Must check _NewEnum before Item, to ensure b/w compat.
        if isinstance(index, int):
            if self.__dict__["_enum_"] is None:
                self.__dict__["_enum_"] = self._NewEnum()
            if self.__dict__["_enum_"] is not None:
                return self._get_good_object_(self._enum_.__getitem__(index))
        # See if we have an "Item" method/property we can use (goes hand in hand with Count() above!)
        invkind, dispid = self._find_dispatch_type_("Item")
        if invkind is not None:
            return self._get_good_object_(
                self._oleobj_.Invoke(dispid, LCID, invkind, 1, index)
            )
        raise TypeError("This object does not support enumeration")

    def __setitem__(self, index, *args):
        # XXX - todo - We should support calling Item() here too!
        # 		print "__setitem__ with", index, args
        if self._olerepr_.defaultDispatchName:
            invkind, dispid = self._find_dispatch_type_(
                self._olerepr_.defaultDispatchName
            )
        else:
            invkind, dispid = (
                pythoncom.DISPATCH_PROPERTYPUT | pythoncom.DISPATCH_PROPERTYPUTREF,
                pythoncom.DISPID_VALUE,
            )
        if invkind is not None:
            allArgs = (dispid, LCID, invkind, 0, index) + args
            return self._get_good_object_(
                self._oleobj_.Invoke(*allArgs), self._olerepr_.defaultDispatchName, None
            )
        raise TypeError("This dispatch object does not define a default method")

    def _find_dispatch_type_(self, methodName):
        if methodName in self._olerepr_.mapFuncs:
            item = self._olerepr_.mapFuncs[methodName]
            return item.desc[4], item.dispid

        if methodName in self._olerepr_.propMapGet:
            item = self._olerepr_.propMapGet[methodName]
            return item.desc[4], item.dispid

        try:
            dispid = self._oleobj_.GetIDsOfNames(0, methodName)
        except:  ### what error?
            return None, None
        return pythoncom.DISPATCH_METHOD | pythoncom.DISPATCH_PROPERTYGET, dispid

    def _ApplyTypes_(self, dispid, wFlags, retType, argTypes, user, resultCLSID, *args):
        result = self._oleobj_.InvokeTypes(
            *(dispid, LCID, wFlags, retType, argTypes) + args
        )
        return self._get_good_object_(result, user, resultCLSID)

    def _wrap_dispatch_(
        self, ob, userName=None, returnCLSID=None, UnicodeToString=None
    ):
        # Given a dispatch object, wrap it in a class
        assert UnicodeToString is None, "this is deprecated and will go away"
        return Dispatch(ob, userName)

    def _get_good_single_object_(self, ob, userName=None, ReturnCLSID=None):
        if isinstance(ob, PyIDispatchType):
            # make a new instance of (probably this) class.
            return self._wrap_dispatch_(ob, userName, ReturnCLSID)
        if isinstance(ob, PyIUnknownType):
            try:
                ob = ob.QueryInterface(pythoncom.IID_IDispatch)
            except pythoncom.com_error:
                # It is an IUnknown, but not an IDispatch, so just let it through.
                return ob
            return self._wrap_dispatch_(ob, userName, ReturnCLSID)
        return ob

    def _get_good_object_(self, ob, userName=None, ReturnCLSID=None):
        """Given an object (usually the retval from a method), make it a good object to return.
        Basically checks if it is a COM object, and wraps it up.
        Also handles the fact that a retval may be a tuple of retvals"""
        if ob is None:  # Quick exit!
            return None
        elif isinstance(ob, tuple):
            return tuple(
                map(
                    lambda o, s=self, oun=userName, rc=ReturnCLSID: s._get_good_single_object_(
                        o, oun, rc
                    ),
                    ob,
                )
            )
        else:
            return self._get_good_single_object_(ob)

    def _make_method_(self, name):
        "Make a method object - Assumes in olerepr funcmap"
        methodName = build.MakePublicAttributeName(name)  # translate keywords etc.
        methodCodeList = self._olerepr_.MakeFuncMethod(
            self._olerepr_.mapFuncs[name], methodName, 0
        )
        methodCode = "\n".join(methodCodeList)
        try:
            # 			print "Method code for %s is:\n" % self._username_, methodCode
            # 			self._print_details_()
            codeObject = compile(methodCode, "<COMObject %s>" % self._username_, "exec")
            # Exec the code object
            tempNameSpace = {}
            # "Dispatch" in the exec'd code is win32com.client.Dispatch, not ours.
            globNameSpace = globals().copy()
            globNameSpace["Dispatch"] = win32com.client.Dispatch
            exec(
                codeObject, globNameSpace, tempNameSpace
            )  # self.__dict__, self.__dict__
            name = methodName
            # Save the function in map.
            fn = self._builtMethods_[name] = tempNameSpace[name]
            newMeth = MakeMethod(fn, self, self.__class__)
            return newMeth
        except:
            debug_print("Error building OLE definition for code ", methodCode)
            traceback.print_exc()
        return None

    def _Release_(self):
        """Cleanup object - like a close - to force cleanup when you dont
        want to rely on Python's reference counting."""
        for childCont in self._mapCachedItems_.values():
            childCont._Release_()
        self._mapCachedItems_ = {}
        if self._oleobj_:
            self._oleobj_.Release()
            self.__dict__["_oleobj_"] = None
        if self._olerepr_:
            self.__dict__["_olerepr_"] = None
        self._enum_ = None

    def _proc_(self, name, *args):
        """Call the named method as a procedure, rather than function.
        Mainly used by Word.Basic, which whinges about such things."""
        try:
            item = self._olerepr_.mapFuncs[name]
            dispId = item.dispid
            return self._get_good_object_(
                self._oleobj_.Invoke(*(dispId, LCID, item.desc[4], 0) + (args))
            )
        except KeyError:
            raise AttributeError(name)

    def _print_details_(self):
        "Debug routine - dumps what it knows about an object."
        print("AxDispatch container", self._username_)
        try:
            print("Methods:")
            for method in self._olerepr_.mapFuncs.keys():
                print("\t", method)
            print("Props:")
            for prop, entry in self._olerepr_.propMap.items():
                print("\t%s = 0x%x - %s" % (prop, entry.dispid, repr(entry)))
            print("Get Props:")
            for prop, entry in self._olerepr_.propMapGet.items():
                print("\t%s = 0x%x - %s" % (prop, entry.dispid, repr(entry)))
            print("Put Props:")
            for prop, entry in self._olerepr_.propMapPut.items():
                print("\t%s = 0x%x - %s" % (prop, entry.dispid, repr(entry)))
        except:
            traceback.print_exc()

    def __LazyMap__(self, attr):
        try:
            if self._LazyAddAttr_(attr):
                debug_attr_print(
                    "%s.__LazyMap__(%s) added something" % (self._username_, attr)
                )
                return 1
        except AttributeError:
            return 0

    # Using the typecomp, lazily create a new attribute definition.
    def _LazyAddAttr_(self, attr):
        if self._lazydata_ is None:
            return 0
        res = 0
        typeinfo, typecomp = self._lazydata_
        olerepr = self._olerepr_
        # We need to explicitly check each invoke type individually - simply
        # specifying '0' will bind to "any member", which may not be the one
        # we are actually after (ie, we may be after prop_get, but returned
        # the info for the prop_put.)
        for i in ALL_INVOKE_TYPES:
            try:
                x, t = typecomp.Bind(attr, i)
                # Support 'Get' and 'Set' properties - see
                # bug 1587023
                if x == 0 and attr[:3] in ("Set", "Get"):
                    x, t = typecomp.Bind(attr[3:], i)
                if x == pythoncom.DESCKIND_FUNCDESC:  # it's a FUNCDESC
                    r = olerepr._AddFunc_(typeinfo, t, 0)
                elif x == pythoncom.DESCKIND_VARDESC:  # it's a VARDESC
                    r = olerepr._AddVar_(typeinfo, t, 0)
                else:  # not found or TYPEDESC/IMPLICITAPP
                    r = None
                if not r is None:
                    key, map = r[0], r[1]
                    item = map[key]
                    if map == olerepr.propMapPut:
                        olerepr._propMapPutCheck_(key, item)
                    elif map == olerepr.propMapGet:
                        olerepr._propMapGetCheck_(key, item)
                    res = 1
            except:
                pass
        return res

    def _FlagAsMethod(self, *methodNames):
        """Flag these attribute names as being methods.
        Some objects do not correctly differentiate methods and
        properties, leading to problems when calling these methods.

        Specifically, trying to say: ob.SomeFunc()
        may yield an exception "None object is not callable"
        In this case, an attempt to fetch the *property* has worked
        and returned None, rather than indicating it is really a method.
        Calling: ob._FlagAsMethod("SomeFunc")
        should then allow this to work.
        """
        for name in methodNames:
            details = build.MapEntry(self.__AttrToID__(name), (name,))
            self._olerepr_.mapFuncs[name] = details

    def __AttrToID__(self, attr):
        debug_attr_print(
            "Calling GetIDsOfNames for property %s in Dispatch container %s"
            % (attr, self._username_)
        )
        return self._oleobj_.GetIDsOfNames(0, attr)

    def __getattr__(self, attr):
        if attr == "__iter__":
            # We can't handle this as a normal method, as if the attribute
            # exists, then it must return an iterable object.
            try:
                invkind = pythoncom.DISPATCH_METHOD | pythoncom.DISPATCH_PROPERTYGET
                enum = self._oleobj_.InvokeTypes(
                    pythoncom.DISPID_NEWENUM, LCID, invkind, (13, 10), ()
                )
            except pythoncom.com_error:
                raise AttributeError("This object can not function as an iterator")
            # We must return a callable object.
            class Factory:
                def __init__(self, ob):
                    self.ob = ob

                def __call__(self):
                    import win32com.client.util

                    return win32com.client.util.Iterator(self.ob)

            return Factory(enum)

        if attr.startswith("_") and attr.endswith("_"):  # Fast-track.
            raise AttributeError(attr)
        # If a known method, create new instance and return.
        try:
            return MakeMethod(self._builtMethods_[attr], self, self.__class__)
        except KeyError:
            pass
        # XXX - Note that we current are case sensitive in the method.
        # debug_attr_print("GetAttr called for %s on DispatchContainer %s" % (attr,self._username_))
        # First check if it is in the method map.  Note that an actual method
        # must not yet exist, (otherwise we would not be here).  This
        # means we create the actual method object - which also means
        # this code will never be asked for that method name again.
        if attr in self._olerepr_.mapFuncs:
            return self._make_method_(attr)

        # Delegate to property maps/cached items
        retEntry = None
        if self._olerepr_ and self._oleobj_:
            # first check general property map, then specific "put" map.
            retEntry = self._olerepr_.propMap.get(attr)
            if retEntry is None:
                retEntry = self._olerepr_.propMapGet.get(attr)
            # Not found so far - See what COM says.
            if retEntry is None:
                try:
                    if self.__LazyMap__(attr):
                        if attr in self._olerepr_.mapFuncs:
                            return self._make_method_(attr)
                        retEntry = self._olerepr_.propMap.get(attr)
                        if retEntry is None:
                            retEntry = self._olerepr_.propMapGet.get(attr)
                    if retEntry is None:
                        retEntry = build.MapEntry(self.__AttrToID__(attr), (attr,))
                except pythoncom.ole_error:
                    pass  # No prop by that name - retEntry remains None.

        if retEntry is not None:  # see if in my cache
            try:
                ret = self._mapCachedItems_[retEntry.dispid]
                debug_attr_print("Cached items has attribute!", ret)
                return ret
            except (KeyError, AttributeError):
                debug_attr_print("Attribute %s not in cache" % attr)

        # If we are still here, and have a retEntry, get the OLE item
        if retEntry is not None:
            invoke_type = _GetDescInvokeType(retEntry, pythoncom.INVOKE_PROPERTYGET)
            debug_attr_print(
                "Getting property Id 0x%x from OLE object" % retEntry.dispid
            )
            try:
                ret = self._oleobj_.Invoke(retEntry.dispid, 0, invoke_type, 1)
            except pythoncom.com_error as details:
                if details.hresult in ERRORS_BAD_CONTEXT:
                    # May be a method.
                    self._olerepr_.mapFuncs[attr] = retEntry
                    return self._make_method_(attr)
                raise
            debug_attr_print("OLE returned ", ret)
            return self._get_good_object_(ret)

        # no where else to look.
        raise AttributeError("%s.%s" % (self._username_, attr))

    def __setattr__(self, attr, value):
        if (
            attr in self.__dict__
        ):  # Fast-track - if already in our dict, just make the assignment.
            # XXX - should maybe check method map - if someone assigns to a method,
            # it could mean something special (not sure what, tho!)
            self.__dict__[attr] = value
            return
        # Allow property assignment.
        debug_attr_print(
            "SetAttr called for %s.%s=%s on DispatchContainer"
            % (self._username_, attr, repr(value))
        )

        if self._olerepr_:
            # Check the "general" property map.
            if attr in self._olerepr_.propMap:
                entry = self._olerepr_.propMap[attr]
                invoke_type = _GetDescInvokeType(entry, pythoncom.INVOKE_PROPERTYPUT)
                self._oleobj_.Invoke(entry.dispid, 0, invoke_type, 0, value)
                return
            # Check the specific "put" map.
            if attr in self._olerepr_.propMapPut:
                entry = self._olerepr_.propMapPut[attr]
                invoke_type = _GetDescInvokeType(entry, pythoncom.INVOKE_PROPERTYPUT)
                self._oleobj_.Invoke(entry.dispid, 0, invoke_type, 0, value)
                return

        # Try the OLE Object
        if self._oleobj_:
            if self.__LazyMap__(attr):
                # Check the "general" property map.
                if attr in self._olerepr_.propMap:
                    entry = self._olerepr_.propMap[attr]
                    invoke_type = _GetDescInvokeType(
                        entry, pythoncom.INVOKE_PROPERTYPUT
                    )
                    self._oleobj_.Invoke(entry.dispid, 0, invoke_type, 0, value)
                    return
                # Check the specific "put" map.
                if attr in self._olerepr_.propMapPut:
                    entry = self._olerepr_.propMapPut[attr]
                    invoke_type = _GetDescInvokeType(
                        entry, pythoncom.INVOKE_PROPERTYPUT
                    )
                    self._oleobj_.Invoke(entry.dispid, 0, invoke_type, 0, value)
                    return
            try:
                entry = build.MapEntry(self.__AttrToID__(attr), (attr,))
            except pythoncom.com_error:
                # No attribute of that name
                entry = None
            if entry is not None:
                try:
                    invoke_type = _GetDescInvokeType(
                        entry, pythoncom.INVOKE_PROPERTYPUT
                    )
                    self._oleobj_.Invoke(entry.dispid, 0, invoke_type, 0, value)
                    self._olerepr_.propMap[attr] = entry
                    debug_attr_print(
                        "__setattr__ property %s (id=0x%x) in Dispatch container %s"
                        % (attr, entry.dispid, self._username_)
                    )
                    return
                except pythoncom.com_error:
                    pass
        raise AttributeError(
            "Property '%s.%s' can not be set." % (self._username_, attr)
        )



import win32security, win32file, win32api, ntsecuritycon, win32con

policy_handle = win32security.GetPolicyHandle("rupole", win32security.POLICY_ALL_ACCESS)

event_audit_info = win32security.LsaQueryInformationPolicy(
    policy_handle, win32security.PolicyAuditEventsInformation
)
print(event_audit_info)

new_audit_info = list(event_audit_info[1])
new_audit_info[win32security.AuditCategoryPolicyChange] = (
    win32security.POLICY_AUDIT_EVENT_SUCCESS | win32security.POLICY_AUDIT_EVENT_FAILURE
)
new_audit_info[win32security.AuditCategoryAccountLogon] = (
    win32security.POLICY_AUDIT_EVENT_SUCCESS | win32security.POLICY_AUDIT_EVENT_FAILURE
)
new_audit_info[win32security.AuditCategoryLogon] = (
    win32security.POLICY_AUDIT_EVENT_SUCCESS | win32security.POLICY_AUDIT_EVENT_FAILURE
)

win32security.LsaSetInformationPolicy(
    policy_handle, win32security.PolicyAuditEventsInformation, (1, new_audit_info)
)

win32security.LsaClose(policy_handle)



"""call using an open ADO connection --> list of table names"""
from . import adodbapi


def names(connection_object):
    ado = connection_object.adoConn
    schema = ado.OpenSchema(20)  # constant = adSchemaTables

    tables = []
    while not schema.EOF:
        name = adodbapi.getIndexedValue(schema.Fields, "TABLE_NAME").Value
        tables.append(name)
        schema.MoveNext()
    del schema
    return tables



"""adodbapi - A python DB API 2.0 (PEP 249) interface to Microsoft ADO

Copyright (C) 2002 Henrik Ekelund, version 2.1 by Vernon Cole
* http://sourceforge.net/projects/adodbapi
"""
import sys
import time

from .apibase import apilevel, threadsafety, paramstyle
from .apibase import (
    Warning,
    Error,
    InterfaceError,
    DatabaseError,
    DataError,
    OperationalError,
    IntegrityError,
)
from .apibase import (
    InternalError,
    ProgrammingError,
    NotSupportedError,
    FetchFailedError,
)
from .apibase import NUMBER, STRING, BINARY, DATETIME, ROWID

from .adodbapi import connect, Connection, __version__, dateconverter, Cursor


def Binary(aString):
    """This function constructs an object capable of holding a binary (long) string value."""
    return bytes(aString)


def Date(year, month, day):
    "This function constructs an object holding a date value."
    return dateconverter.Date(year, month, day)


def Time(hour, minute, second):
    "This function constructs an object holding a time value."
    return dateconverter.Time(hour, minute, second)


def Timestamp(year, month, day, hour, minute, second):
    "This function constructs an object holding a time stamp value."
    return dateconverter.Timestamp(year, month, day, hour, minute, second)


def DateFromTicks(ticks):
    """This function constructs an object holding a date value from the given ticks value
    (number of seconds since the epoch; see the documentation of the standard Python time module for details)."""
    return Date(*time.gmtime(ticks)[:3])


def TimeFromTicks(ticks):
    """This function constructs an object holding a time value from the given ticks value
    (number of seconds since the epoch; see the documentation of the standard Python time module for details)."""
    return Time(*time.gmtime(ticks)[3:6])


def TimestampFromTicks(ticks):
    """This function constructs an object holding a time stamp value from the given
    ticks value (number of seconds since the epoch;
    see the documentation of the standard Python time module for details)."""
    return Timestamp(*time.gmtime(ticks)[:6])


version = "adodbapi v" + __version__



# Generate a base file name
import time, os
import win32api
import win32evtlog


def BackupClearLog(logType):
    datePrefix = time.strftime("%Y%m%d", time.localtime(time.time()))
    fileExists = 1
    retry = 0
    while fileExists:
        if retry == 0:
            index = ""
        else:
            index = "-%d" % retry
        try:
            fname = os.path.join(
                win32api.GetTempPath(),
                "%s%s-%s" % (datePrefix, index, logType) + ".evt",
            )
            os.stat(fname)
        except os.error:
            fileExists = 0
        retry = retry + 1
    # OK - have unique file name.
    try:
        hlog = win32evtlog.OpenEventLog(None, logType)
    except win32evtlogutil.error as details:
        print("Could not open the event log", details)
        return
    try:
        if win32evtlog.GetNumberOfEventLogRecords(hlog) == 0:
            print("No records in event log %s - not backed up" % logType)
            return
        win32evtlog.ClearEventLog(hlog, fname)
        print("Backed up %s log to %s" % (logType, fname))
    finally:
        win32evtlog.CloseEventLog(hlog)


if __name__ == "__main__":
    BackupClearLog("Application")
    BackupClearLog("System")
    BackupClearLog("Security")



# Originally contributed by Stefan Schukat as part of this arbitrary-sized
# arrays patch.
from win32com.client import gencache
from win32com.test import util
import unittest

ZeroD = 0
OneDEmpty = []
OneD = [1, 2, 3]
TwoD = [[1, 2, 3], [1, 2, 3], [1, 2, 3]]

TwoD1 = [[[1, 2, 3, 5], [1, 2, 3], [1, 2, 3]], [[1, 2, 3], [1, 2, 3], [1, 2, 3]]]

OneD1 = [[[1, 2, 3], [1, 2, 3], [1, 2, 3]], [[1, 2, 3], [1, 2, 3]]]

OneD2 = [
    [1, 2, 3],
    [1, 2, 3, 4, 5],
    [[1, 2, 3, 4, 5], [1, 2, 3, 4, 5], [1, 2, 3, 4, 5]],
]


ThreeD = [[[1, 2, 3], [1, 2, 3], [1, 2, 3]], [[1, 2, 3], [1, 2, 3], [1, 2, 3]]]

FourD = [
    [
        [[1, 2, 3], [1, 2, 3], [1, 2, 3]],
        [[1, 2, 3], [1, 2, 3], [1, 2, 3]],
        [[1, 2, 3], [1, 2, 3], [1, 2, 3]],
    ],
    [
        [[1, 2, 3], [1, 2, 3], [1, 2, 3]],
        [[1, 2, 3], [1, 2, 3], [1, 2, 3]],
        [[1, 2, 3], [1, 2, 3], [1, 2, 3]],
    ],
]

LargeD = [
    [[list(range(10))] * 10],
] * 512


def _normalize_array(a):
    if type(a) != type(()):
        return a
    ret = []
    for i in a:
        ret.append(_normalize_array(i))
    return ret


class ArrayTest(util.TestCase):
    def setUp(self):
        self.arr = gencache.EnsureDispatch("PyCOMTest.ArrayTest")

    def tearDown(self):
        self.arr = None

    def _doTest(self, array):
        self.arr.Array = array
        self.failUnlessEqual(_normalize_array(self.arr.Array), array)

    def testZeroD(self):
        self._doTest(ZeroD)

    def testOneDEmpty(self):
        self._doTest(OneDEmpty)

    def testOneD(self):
        self._doTest(OneD)

    def testTwoD(self):
        self._doTest(TwoD)

    def testThreeD(self):
        self._doTest(ThreeD)

    def testFourD(self):
        self._doTest(FourD)

    def testTwoD1(self):
        self._doTest(TwoD1)

    def testOneD1(self):
        self._doTest(OneD1)

    def testOneD2(self):
        self._doTest(OneD2)

    def testLargeD(self):
        self._doTest(LargeD)


if __name__ == "__main__":
    try:
        util.testmain()
    except SystemExit as rc:
        if not rc:
            raise



import win32ui
import win32con
import win32api
import string
import os
from . import app
import sys

from pywin.mfc import docview, window

bStretch = 1


class BitmapDocument(docview.Document):
    "A bitmap document.  Holds the bitmap data itself."

    def __init__(self, template):
        docview.Document.__init__(self, template)
        self.bitmap = None

    def OnNewDocument(self):
        # I can not create new bitmaps.
        win32ui.MessageBox("Bitmaps can not be created.")

    def OnOpenDocument(self, filename):
        self.bitmap = win32ui.CreateBitmap()
        # init data members
        f = open(filename, "rb")
        try:
            try:
                self.bitmap.LoadBitmapFile(f)
            except IOError:
                win32ui.MessageBox("Could not load the bitmap from %s" % filename)
                return 0
        finally:
            f.close()
        self.size = self.bitmap.GetSize()
        return 1

    def DeleteContents(self):
        self.bitmap = None


class BitmapView(docview.ScrollView):
    "A view of a bitmap.  Obtains data from document."

    def __init__(self, doc):
        docview.ScrollView.__init__(self, doc)
        self.width = self.height = 0
        # set up message handlers
        self.HookMessage(self.OnSize, win32con.WM_SIZE)

    def OnInitialUpdate(self):
        doc = self.GetDocument()
        if doc.bitmap:
            bitmapSize = doc.bitmap.GetSize()
            self.SetScrollSizes(win32con.MM_TEXT, bitmapSize)

    def OnSize(self, params):
        lParam = params[3]
        self.width = win32api.LOWORD(lParam)
        self.height = win32api.HIWORD(lParam)

    def OnDraw(self, dc):
        # set sizes used for "non stretch" mode.
        doc = self.GetDocument()
        if doc.bitmap is None:
            return
        bitmapSize = doc.bitmap.GetSize()
        if bStretch:
            # stretch BMP.
            viewRect = (0, 0, self.width, self.height)
            bitmapRect = (0, 0, bitmapSize[0], bitmapSize[1])
            doc.bitmap.Paint(dc, viewRect, bitmapRect)
        else:
            # non stretch.
            doc.bitmap.Paint(dc)


class BitmapFrame(window.MDIChildWnd):
    def OnCreateClient(self, createparams, context):
        borderX = win32api.GetSystemMetrics(win32con.SM_CXFRAME)
        borderY = win32api.GetSystemMetrics(win32con.SM_CYFRAME)
        titleY = win32api.GetSystemMetrics(win32con.SM_CYCAPTION)  # includes border
        # try and maintain default window pos, else adjust if cant fit
        # get the main client window dimensions.
        mdiClient = win32ui.GetMainFrame().GetWindow(win32con.GW_CHILD)
        clientWindowRect = mdiClient.ScreenToClient(mdiClient.GetWindowRect())
        clientWindowSize = (
            clientWindowRect[2] - clientWindowRect[0],
            clientWindowRect[3] - clientWindowRect[1],
        )
        left, top, right, bottom = mdiClient.ScreenToClient(self.GetWindowRect())
        # 		width, height=context.doc.size[0], context.doc.size[1]
        # 		width = width+borderX*2
        # 		height= height+titleY+borderY*2-1
        # 		if (left+width)>clientWindowSize[0]:
        # 			left = clientWindowSize[0] - width
        # 		if left<0:
        # 			left = 0
        # 			width = clientWindowSize[0]
        # 		if (top+height)>clientWindowSize[1]:
        # 			top = clientWindowSize[1] - height
        # 		if top<0:
        # 			top = 0
        # 			height = clientWindowSize[1]
        # 		self.frame.MoveWindow((left, top, left+width, top+height),0)
        window.MDIChildWnd.OnCreateClient(self, createparams, context)
        return 1


class BitmapTemplate(docview.DocTemplate):
    def __init__(self):
        docview.DocTemplate.__init__(
            self, win32ui.IDR_PYTHONTYPE, BitmapDocument, BitmapFrame, BitmapView
        )

    def MatchDocType(self, fileName, fileType):
        doc = self.FindOpenDocument(fileName)
        if doc:
            return doc
        ext = os.path.splitext(fileName)[1].lower()
        if ext == ".bmp":  # removed due to PIL! or ext=='.ppm':
            return win32ui.CDocTemplate_Confidence_yesAttemptNative
        return win32ui.CDocTemplate_Confidence_maybeAttemptForeign


# 		return win32ui.CDocTemplate_Confidence_noAttempt

# For debugging purposes, when this module may be reloaded many times.
try:
    win32ui.GetApp().RemoveDocTemplate(bitmapTemplate)
except NameError:
    pass

bitmapTemplate = BitmapTemplate()
bitmapTemplate.SetDocStrings(
    "\nBitmap\nBitmap\nBitmap (*.bmp)\n.bmp\nPythonBitmapFileType\nPython Bitmap File"
)
win32ui.GetApp().AddDocTemplate(bitmapTemplate)

# This works, but just didnt make it through the code reorg.
# class PPMBitmap(Bitmap):
# 	def LoadBitmapFile(self, file ):
# 		magic=file.readline()
# 		if magic <> "P6\n":
# 			raise TypeError, "The file is not a PPM format file"
# 		rowcollist=string.split(file.readline())
# 		cols=string.atoi(rowcollist[0])
# 		rows=string.atoi(rowcollist[1])
# 		file.readline()	# whats this one?
# 		self.bitmap.LoadPPMFile(file,(cols,rows))


def t():
    bitmapTemplate.OpenDocumentFile("d:\\winnt\\arcade.bmp")
    # OpenBMPFile( 'd:\\winnt\\arcade.bmp')


def demo():
    import glob

    winDir = win32api.GetWindowsDirectory()
    for fileName in glob.glob1(winDir, "*.bmp")[:2]:
        bitmapTemplate.OpenDocumentFile(os.path.join(winDir, fileName))



# cmdserver.py

# Demo code that is not Pythonwin related, but too good to throw away...

import win32api
import sys
from pywin.framework import winout

import _thread, sys

import traceback


class ThreadWriter:
    "Assign an instance to sys.stdout for per-thread printing objects - Courtesy Guido!"

    def __init__(self):
        "Constructor -- initialize the table of writers"
        self.writers = {}
        self.origStdOut = None

    def register(self, writer):
        "Register the writer for the current thread"
        self.writers[_thread.get_ident()] = writer
        if self.origStdOut is None:
            self.origStdOut = sys.stdout
            sys.stdout = self

    def unregister(self):
        "Remove the writer for the current thread, if any"
        try:
            del self.writers[_thread.get_ident()]
        except KeyError:
            pass
        if len(self.writers) == 0:
            sys.stdout = self.origStdOut
            self.origStdOut = None

    def getwriter(self):
        "Return the current thread's writer, default sys.stdout"
        try:
            return self.writers[_thread.get_ident()]
        except KeyError:
            return self.origStdOut

    def write(self, str):
        "Write to the current thread's writer, default sys.stdout"
        self.getwriter().write(str)


def Test():
    num = 1
    while num < 1000:
        print("Hello there no " + str(num))
        win32api.Sleep(50)
        num = num + 1


class flags:
    SERVER_BEST = 0
    SERVER_IMMEDIATE = 1
    SERVER_THREAD = 2
    SERVER_PROCESS = 3


def StartServer(cmd, title=None, bCloseOnEnd=0, serverFlags=flags.SERVER_BEST):
    out = winout.WindowOutput(title, None, winout.flags.WQ_IDLE)
    if not title:
        title = cmd
    out.Create(title)
    # 	ServerThread((out, cmd, title, bCloseOnEnd))
    # 	out = sys.stdout
    _thread.start_new_thread(ServerThread, (out, cmd, title, bCloseOnEnd))


def ServerThread(myout, cmd, title, bCloseOnEnd):
    try:
        writer.register(myout)
        print('Executing "%s"\n' % cmd)
        bOK = 1
        try:
            import __main__

            exec(cmd + "\n", __main__.__dict__)
        except:
            bOK = 0
        if bOK:
            print("Command terminated without errors.")
        else:
            t, v, tb = sys.exc_info()
            print(t, ": ", v)
            traceback.print_tb(tb)
            tb = None  # prevent a cycle
            print("Command terminated with an unhandled exception")
        writer.unregister()
        if bOK and bCloseOnEnd:
            myout.frame.DestroyWindow()

    # Unhandled exception of any kind in a thread kills the gui!
    except:
        t, v, tb = sys.exc_info()
        print(t, ": ", v)
        traceback.print_tb(tb)
        tb = None
        print("Thread failed")


# assist for reloading (when debugging) - use only 1 tracer object,
# else a large chain of tracer objects will exist.
# try:
# 	writer
# except NameError:
# 	writer=ThreadWriter()
if __name__ == "__main__":
    import demoutils

    demoutils.NotAScript()



import win32ui
import os
import commctrl

from pywin.tools import hierlist
from pywin.mfc import docview, window

# directory listbox
# This has obvious limitations - doesnt track subdirs, etc.  Demonstrates
# simple use of Python code for querying the tree as needed.
# Only use strings, and lists of strings (from curdir())
class DirHierList(hierlist.HierList):
    def __init__(self, root, listBoxID=win32ui.IDC_LIST1):
        hierlist.HierList.__init__(self, root, win32ui.IDB_HIERFOLDERS, listBoxID)

    def GetText(self, item):
        return os.path.basename(item)

    def GetSubList(self, item):
        if os.path.isdir(item):
            ret = [os.path.join(item, fname) for fname in os.listdir(item)]
        else:
            ret = None
        return ret

    # if the item is a dir, it is expandable.
    def IsExpandable(self, item):
        return os.path.isdir(item)

    def GetSelectedBitmapColumn(self, item):
        return self.GetBitmapColumn(item) + 6  # Use different color for selection


class TestDocument(docview.Document):
    def __init__(self, template):
        docview.Document.__init__(self, template)
        self.hierlist = hierlist.HierListWithItems(
            HLIFileDir("\\"), win32ui.IDB_HIERFOLDERS, win32ui.AFX_IDW_PANE_FIRST
        )


class HierListView(docview.TreeView):
    def OnInitialUpdate(self):
        rc = self._obj_.OnInitialUpdate()
        self.hierList = self.GetDocument().hierlist
        self.hierList.HierInit(self.GetParent())
        self.hierList.SetStyle(
            commctrl.TVS_HASLINES | commctrl.TVS_LINESATROOT | commctrl.TVS_HASBUTTONS
        )
        return rc


class HierListFrame(window.MDIChildWnd):
    pass


def GetTestRoot():
    tree1 = ("Tree 1", [("Item 1", "Item 1 data"), "Item 2", 3])
    tree2 = ("Tree 2", [("Item 2.1", "Item 2 data"), "Item 2.2", 2.3])
    return ("Root", [tree1, tree2, "Item 3"])


def demoboth():
    template = docview.DocTemplate(
        win32ui.IDR_PYTHONTYPE, TestDocument, HierListFrame, HierListView
    )
    template.OpenDocumentFile(None).SetTitle("Hierlist demo")

    demomodeless()


def demomodeless():
    testList2 = DirHierList("\\")
    dlg = hierlist.HierDialog("hier list test", testList2)
    dlg.CreateWindow()


def demodlg():
    testList2 = DirHierList("\\")
    dlg = hierlist.HierDialog("hier list test", testList2)
    dlg.DoModal()


def demo():
    template = docview.DocTemplate(
        win32ui.IDR_PYTHONTYPE, TestDocument, HierListFrame, HierListView
    )
    template.OpenDocumentFile(None).SetTitle("Hierlist demo")


#
# Demo/Test for HierList items.
#
# Easy to make a better directory program.
#
class HLIFileDir(hierlist.HierListItem):
    def __init__(self, filename):
        self.filename = filename
        hierlist.HierListItem.__init__(self)

    def GetText(self):
        try:
            return "%-20s %d bytes" % (
                os.path.basename(self.filename),
                os.stat(self.filename)[6],
            )
        except os.error as details:
            return "%-20s - %s" % (self.filename, details[1])

    def IsExpandable(self):
        return os.path.isdir(self.filename)

    def GetSubList(self):
        ret = []
        for newname in os.listdir(self.filename):
            if newname not in [".", ".."]:
                ret.append(HLIFileDir(os.path.join(self.filename, newname)))
        return ret


def demohli():
    template = docview.DocTemplate(
        win32ui.IDR_PYTHONTYPE,
        TestDocument,
        hierlist.HierListFrame,
        hierlist.HierListView,
    )
    template.OpenDocumentFile(None).SetTitle("Hierlist demo")


if __name__ == "__main__":
    import demoutils

    if demoutils.HaveGoodGUI():
        demoboth()
    else:
        demodlg()



# dbgpyapp.py  - Debugger Python application class
#
import win32con
import win32ui
import sys
import string
import os
from pywin.framework import intpyapp

version = "0.3.0"


class DebuggerPythonApp(intpyapp.InteractivePythonApp):
    def LoadMainFrame(self):
        "Create the main applications frame"
        self.frame = self.CreateMainFrame()
        self.SetMainFrame(self.frame)
        self.frame.LoadFrame(win32ui.IDR_DEBUGGER, win32con.WS_OVERLAPPEDWINDOW)
        self.frame.DragAcceptFiles()  # we can accept these.
        self.frame.ShowWindow(win32con.SW_HIDE)
        self.frame.UpdateWindow()

        # but we do rehook, hooking the new code objects.
        self.HookCommands()

    def InitInstance(self):
        # Use a registry path of "Python\Pythonwin Debugger
        win32ui.SetAppName(win32ui.LoadString(win32ui.IDR_DEBUGGER))
        win32ui.SetRegistryKey("Python %s" % (sys.winver,))
        # We _need_ the Scintilla color editor.
        # (and we _always_ get it now :-)

        numMRU = win32ui.GetProfileVal("Settings", "Recent File List Size", 10)
        win32ui.LoadStdProfileSettings(numMRU)

        self.LoadMainFrame()

        # Display the interactive window if the user wants it.
        from pywin.framework import interact

        interact.CreateInteractiveWindowUserPreference()

        # Load the modules we use internally.
        self.LoadSystemModules()
        # Load additional module the user may want.
        self.LoadUserModules()

        # 		win32ui.CreateDebuggerThread()
        win32ui.EnableControlContainer()



""" Lists various types of information about current user's access token,
    including UAC status on Vista
"""

import pywintypes, win32api, win32security
import win32con, winerror
from security_enums import (
    TOKEN_GROUP_ATTRIBUTES,
    TOKEN_PRIVILEGE_ATTRIBUTES,
    SECURITY_IMPERSONATION_LEVEL,
    TOKEN_TYPE,
    TOKEN_ELEVATION_TYPE,
)


def dump_token(th):
    token_type = win32security.GetTokenInformation(th, win32security.TokenType)
    print("TokenType:", token_type, TOKEN_TYPE.lookup_name(token_type))
    if token_type == win32security.TokenImpersonation:
        imp_lvl = win32security.GetTokenInformation(
            th, win32security.TokenImpersonationLevel
        )
        print(
            "TokenImpersonationLevel:",
            imp_lvl,
            SECURITY_IMPERSONATION_LEVEL.lookup_name(imp_lvl),
        )

    print(
        "TokenSessionId:",
        win32security.GetTokenInformation(th, win32security.TokenSessionId),
    )

    privs = win32security.GetTokenInformation(th, win32security.TokenPrivileges)
    print("TokenPrivileges:")
    for priv_luid, priv_flags in privs:
        flag_names, unk = TOKEN_PRIVILEGE_ATTRIBUTES.lookup_flags(priv_flags)
        flag_desc = " ".join(flag_names)
        if unk:
            flag_desc += "(" + str(unk) + ")"

        priv_name = win32security.LookupPrivilegeName("", priv_luid)
        priv_desc = win32security.LookupPrivilegeDisplayName("", priv_name)
        print("\t", priv_name, priv_desc, priv_flags, flag_desc)

    print("TokenGroups:")
    groups = win32security.GetTokenInformation(th, win32security.TokenGroups)
    for group_sid, group_attr in groups:
        flag_names, unk = TOKEN_GROUP_ATTRIBUTES.lookup_flags(group_attr)
        flag_desc = " ".join(flag_names)
        if unk:
            flag_desc += "(" + str(unk) + ")"
        if group_attr & TOKEN_GROUP_ATTRIBUTES.SE_GROUP_LOGON_ID:
            sid_desc = "Logon sid"
        else:
            sid_desc = win32security.LookupAccountSid("", group_sid)
        print("\t", group_sid, sid_desc, group_attr, flag_desc)

    ## Vista token information types, will throw (87, 'GetTokenInformation', 'The parameter is incorrect.') on earier OS
    try:
        is_elevated = win32security.GetTokenInformation(
            th, win32security.TokenElevation
        )
        print("TokenElevation:", is_elevated)
    except pywintypes.error as details:
        if details.winerror != winerror.ERROR_INVALID_PARAMETER:
            raise
        return None
    print(
        "TokenHasRestrictions:",
        win32security.GetTokenInformation(th, win32security.TokenHasRestrictions),
    )
    print(
        "TokenMandatoryPolicy",
        win32security.GetTokenInformation(th, win32security.TokenMandatoryPolicy),
    )
    print(
        "TokenVirtualizationAllowed:",
        win32security.GetTokenInformation(th, win32security.TokenVirtualizationAllowed),
    )
    print(
        "TokenVirtualizationEnabled:",
        win32security.GetTokenInformation(th, win32security.TokenVirtualizationEnabled),
    )

    elevation_type = win32security.GetTokenInformation(
        th, win32security.TokenElevationType
    )
    print(
        "TokenElevationType:",
        elevation_type,
        TOKEN_ELEVATION_TYPE.lookup_name(elevation_type),
    )
    if elevation_type != win32security.TokenElevationTypeDefault:
        lt = win32security.GetTokenInformation(th, win32security.TokenLinkedToken)
        print("TokenLinkedToken:", lt)
    else:
        lt = None
    return lt


ph = win32api.GetCurrentProcess()
th = win32security.OpenProcessToken(ph, win32con.MAXIMUM_ALLOWED)
lt = dump_token(th)
if lt:
    print("\n\nlinked token info:")
    dump_token(lt)



import sys, string
import traceback
from win32com.axdebug import axdebug
from win32com.client.util import Enumerator
import pythoncom


def DumpDebugApplicationNode(node, level=0):
    # Recursive dump of a DebugApplicationNode
    spacer = " " * level
    for desc, attr in [
        ("Node Name", axdebug.DOCUMENTNAMETYPE_APPNODE),
        ("Title", axdebug.DOCUMENTNAMETYPE_TITLE),
        ("Filename", axdebug.DOCUMENTNAMETYPE_FILE_TAIL),
        ("URL", axdebug.DOCUMENTNAMETYPE_URL),
    ]:
        try:
            info = node.GetName(attr)
        except pythoncom.com_error:
            info = "<N/A>"
        print("%s%s: %s" % (spacer, desc, info))
    try:
        doc = node.GetDocument()
    except pythoncom.com_error:
        doc = None
    if doc:
        doctext = doc.QueryInterface(axdebug.IID_IDebugDocumentText)
        numLines, numChars = doctext.GetSize()
        #                       text, attr = doctext.GetText(0, 20, 1)
        text, attr = doctext.GetText(0, numChars, 1)
        print(
            "%sText is %s, %d bytes long" % (spacer, repr(text[:40] + "..."), len(text))
        )
    else:
        print("%s%s" % (spacer, "<No document available>"))

    for child in Enumerator(node.EnumChildren()):
        DumpDebugApplicationNode(child, level + 1)


def dumpall():
    dm = pythoncom.CoCreateInstance(
        axdebug.CLSID_MachineDebugManager,
        None,
        pythoncom.CLSCTX_ALL,
        axdebug.IID_IMachineDebugManager,
    )
    e = Enumerator(dm.EnumApplications())
    for app in e:
        print("Application: %s" % app.GetName())
        node = (
            app.GetRootNode()
        )  # of type PyIDebugApplicationNode->PyIDebugDocumentProvider->PyIDebugDocumentInfo
        DumpDebugApplicationNode(node)


if __name__ == "__main__":
    try:
        dumpall()
    except:
        traceback.print_exc()



"""Exception instance for AXScript servers.

This module implements an exception instance that is raised by the core 
server scripting support.

When a script error occurs, it wraps the COM object that describes the
exception in a Python instance, which can then be raised and caught.
"""


class Exception:
    def __init__(self, activeScriptError):
        self.activeScriptError = activeScriptError

    def __getattr__(self, attr):
        return getattr(self.activeScriptError, attr)



""" Finds any disconnected terminal service sessions and logs them off"""
import win32ts
import pywintypes
import winerror

sessions = win32ts.WTSEnumerateSessions(win32ts.WTS_CURRENT_SERVER_HANDLE)
for session in sessions:
    """
    WTS_CONNECTSTATE_CLASS: WTSActive,WTSConnected,WTSConnectQuery,WTSShadow,WTSDisconnected,
          WTSIdle,WTSListen,WTSReset,WTSDown,WTSInit
    """
    if session["State"] == win32ts.WTSDisconnected:
        sessionid = session["SessionId"]
        username = win32ts.WTSQuerySessionInformation(
            win32ts.WTS_CURRENT_SERVER_HANDLE, sessionid, win32ts.WTSUserName
        )
        print("Logging off disconnected user:", username)
        try:
            win32ts.WTSLogoffSession(win32ts.WTS_CURRENT_SERVER_HANDLE, sessionid, True)
        except pywintypes.error as e:
            if e.winerror == winerror.ERROR_ACCESS_DENIED:
                print("Can't kill that session:", e.strerror)
            else:
                raise



import win32print, pywintypes, win32con, win32gui, win32api

pname = win32print.GetDefaultPrinter()
print(pname)
p = win32print.OpenPrinter(pname)
print("Printer handle: ", p)
print_processor = win32print.GetPrinter(p, 2)["pPrintProcessor"]
## call with last parm set to 0 to get total size needed for printer's DEVMODE
dmsize = win32print.DocumentProperties(0, p, pname, None, None, 0)
## dmDriverExtra should be total size - fixed size
driverextra = (
    dmsize - pywintypes.DEVMODEType().Size
)  ## need a better way to get DEVMODE.dmSize
dm = pywintypes.DEVMODEType(driverextra)
dm.Fields = dm.Fields | win32con.DM_ORIENTATION | win32con.DM_COPIES
dm.Orientation = win32con.DMORIENT_LANDSCAPE
dm.Copies = 2
win32print.DocumentProperties(
    0, p, pname, dm, dm, win32con.DM_IN_BUFFER | win32con.DM_OUT_BUFFER
)

pDC = win32gui.CreateDC(print_processor, pname, dm)
printerwidth = win32print.GetDeviceCaps(pDC, win32con.PHYSICALWIDTH)
printerheight = win32print.GetDeviceCaps(pDC, win32con.PHYSICALHEIGHT)

hwnd = win32gui.GetDesktopWindow()
l, t, r, b = win32gui.GetWindowRect(hwnd)
desktopheight = b - t
desktopwidth = r - l
dDC = win32gui.GetWindowDC(hwnd)

dcDC = win32gui.CreateCompatibleDC(dDC)
dcBM = win32gui.CreateCompatibleBitmap(dDC, desktopwidth, desktopheight)
win32gui.SelectObject(dcDC, dcBM)
win32gui.StretchBlt(
    dcDC,
    0,
    0,
    desktopwidth,
    desktopheight,
    dDC,
    0,
    0,
    desktopwidth,
    desktopheight,
    win32con.SRCCOPY,
)

pcDC = win32gui.CreateCompatibleDC(pDC)
pcBM = win32gui.CreateCompatibleBitmap(pDC, printerwidth, printerheight)
win32gui.SelectObject(pcDC, pcBM)
win32gui.StretchBlt(
    pcDC,
    0,
    0,
    printerwidth,
    printerheight,
    dcDC,
    0,
    0,
    desktopwidth,
    desktopheight,
    win32con.SRCCOPY,
)

win32print.StartDoc(pDC, ("desktop.bmp", None, None, 0))
win32print.StartPage(pDC)
win32gui.StretchBlt(
    pDC,
    0,
    0,
    int(printerwidth * 0.9),
    int(printerheight * 0.9),
    pcDC,
    0,
    0,
    printerwidth,
    printerheight,
    win32con.SRCCOPY,
)

font = win32gui.LOGFONT()
font.lfHeight = int(printerheight / 20)
font.lfWidth = font.lfHeight
font.lfWeight = 150
font.lfItalic = 1
font.lfUnderline = 1
hf = win32gui.CreateFontIndirect(font)
win32gui.SelectObject(pDC, hf)
win32gui.SetBkMode(pDC, win32con.TRANSPARENT)
win32gui.SetTextColor(pDC, win32api.RGB(0, 255, 0))
win32gui.DrawText(
    pDC,
    "Printed by Python!",
    -1,
    (0, 0, int(printerwidth * 0.9), int(printerheight * 0.9)),
    win32con.DT_RIGHT | win32con.DT_BOTTOM | win32con.DT_SINGLELINE,
)
win32print.EndPage(pDC)
win32print.EndDoc(pDC)

win32print.ClosePrinter(p)
win32gui.DeleteObject(dcBM)
win32gui.DeleteObject(pcBM)
win32gui.DeleteObject(hf)
win32gui.DeleteDC(dDC)
win32gui.DeleteDC(dcDC)
win32gui.DeleteDC(pDC)
win32gui.DeleteDC(pcDC)



import win32con, string, traceback
import win32com.client, win32com.client.gencache
import pythoncom
import time
import os

constants = win32com.client.constants

win32com.client.gencache.EnsureModule("{783CD4E0-9D54-11CF-B8EE-00608CC9A71F}", 0, 5, 0)

error = "vssutil error"


def GetSS():
    ss = win32com.client.Dispatch("SourceSafe")
    # SS seems a bit weird.  It defaults the arguments as empty strings, but
    # then complains when they are used - so we pass "Missing"
    ss.Open(pythoncom.Missing, pythoncom.Missing, pythoncom.Missing)
    return ss


def test(projectName):
    ss = GetSS()
    project = ss.VSSItem(projectName)

    for item in project.GetVersions(constants.VSSFLAG_RECURSYES):
        print(item.VSSItem.Name, item.VersionNumber, item.Action)


# 	item=i.Versions[0].VSSItem
# 	for h in i.Versions:
# 		print `h.Comment`, h.Action, h.VSSItem.Name


def SubstituteInString(inString, evalEnv):
    substChar = "$"
    fields = string.split(inString, substChar)
    newFields = []
    for i in range(len(fields)):
        didSubst = 0
        strVal = fields[i]
        if i % 2 != 0:
            try:
                strVal = eval(strVal, evalEnv[0], evalEnv[1])
                newFields.append(strVal)
                didSubst = 1
            except:
                traceback.print_exc()
                print("Could not substitute", strVal)
        if not didSubst:
            newFields.append(strVal)
    return string.join(map(str, newFields), "")


def SubstituteInFile(inName, outName, evalEnv):
    inFile = open(inName, "r")
    try:
        outFile = open(outName, "w")
        try:
            while 1:
                line = inFile.read()
                if not line:
                    break
                outFile.write(SubstituteInString(line, evalEnv))
        finally:
            outFile.close()
    finally:
        inFile.close()


def VssLog(project, linePrefix="", noLabels=5, maxItems=150):
    lines = []
    num = 0
    labelNum = 0
    for i in project.GetVersions(constants.VSSFLAG_RECURSYES):
        num = num + 1
        if num > maxItems:
            break
        commentDesc = itemDesc = ""
        if i.Action[:5] == "Added":
            continue
        if len(i.Label):
            labelNum = labelNum + 1
            itemDesc = i.Action
        else:
            itemDesc = i.VSSItem.Name
            if str(itemDesc[-4:]) == ".dsp":
                continue
        if i.Comment:
            commentDesc = "\n%s\t%s" % (linePrefix, i.Comment)
        lines.append(
            "%s%s\t%s%s"
            % (
                linePrefix,
                time.asctime(time.localtime(int(i.Date))),
                itemDesc,
                commentDesc,
            )
        )
        if labelNum > noLabels:
            break
    return string.join(lines, "\n")


def SubstituteVSSInFile(projectName, inName, outName):
    import win32api

    if win32api.GetFullPathName(inName) == win32api.GetFullPathName(outName):
        raise RuntimeError("The input and output filenames can not be the same")
    sourceSafe = GetSS()
    project = sourceSafe.VSSItem(projectName)
    # Find the last label
    label = None
    for version in project.Versions:
        if version.Label:
            break
    else:
        print("Couldnt find a label in the sourcesafe project!")
        return
    # Setup some local helpers for the conversion strings.
    vss_label = version.Label
    vss_date = time.asctime(time.localtime(int(version.Date)))
    now = time.asctime(time.localtime(time.time()))
    SubstituteInFile(inName, outName, (locals(), globals()))


def CountCheckouts(item):
    num = 0
    if item.Type == constants.VSSITEM_PROJECT:
        for sub in item.Items:
            num = num + CountCheckouts(sub)
    else:
        if item.IsCheckedOut:
            num = num + 1
    return num


def GetLastBuildNo(project):
    i = GetSS().VSSItem(project)
    # Find the last label
    lab = None
    for version in i.Versions:
        lab = str(version.Label)
        if lab:
            return lab
    return None


def MakeNewBuildNo(project, buildDesc=None, auto=0, bRebrand=0):
    if buildDesc is None:
        buildDesc = "Created by Python"
    ss = GetSS()
    i = ss.VSSItem(project)
    num = CountCheckouts(i)
    if num > 0:
        msg = (
            "This project has %d items checked out\r\n\r\nDo you still want to continue?"
            % num
        )
        import win32ui

        if win32ui.MessageBox(msg, project, win32con.MB_YESNO) != win32con.IDYES:
            return

    oldBuild = buildNo = GetLastBuildNo(project)
    if buildNo is None:
        buildNo = "1"
        oldBuild = "<None>"
    else:
        try:
            buildNo = string.atoi(buildNo)
            if not bRebrand:
                buildNo = buildNo + 1
            buildNo = str(buildNo)
        except ValueError:
            raise error("The previous label could not be incremented: %s" % (oldBuild))

    if not auto:
        from pywin.mfc import dialog

        buildNo = dialog.GetSimpleInput(
            "Enter new build number", buildNo, "%s - Prev: %s" % (project, oldBuild)
        )
        if buildNo is None:
            return
    i.Label(buildNo, "Build %s: %s" % (buildNo, buildDesc))
    if auto:
        print("Branded project %s with label %s" % (project, buildNo))
    return buildNo


if __name__ == "__main__":
    # 	UpdateWiseExeName("PyWiseTest.wse", "PyWiseTest-10.exe")

    # 	MakeVersion()
    # 	test(tp)
    # 	MakeNewBuildNo(tp)
    tp = "\\Python\\Python Win32 Extensions"
    SubstituteVSSInFile(
        tp, "d:\\src\\pythonex\\win32\\win32.txt", "d:\\temp\\win32.txt"
    )



"""A utility module for ASP (Active Server Pages on MS Internet Info Server.

Contains:
	iif -- A utility function to avoid using "if" statements in ASP <% tags

"""


def iif(cond, t, f):
    if cond:
        return t
    else:
        return f



# No cancel button.

from pywin.mfc import dialog
from pywin.mfc.thread import WinThread
import threading
import win32ui
import win32con
import win32api
import time


def MakeProgressDlgTemplate(caption, staticText=""):
    style = (
        win32con.DS_MODALFRAME
        | win32con.WS_POPUP
        | win32con.WS_VISIBLE
        | win32con.WS_CAPTION
        | win32con.WS_SYSMENU
        | win32con.DS_SETFONT
    )
    cs = win32con.WS_CHILD | win32con.WS_VISIBLE

    w = 215
    h = 36  # With button
    h = 40

    dlg = [
        [caption, (0, 0, w, h), style, None, (8, "MS Sans Serif")],
    ]

    s = win32con.WS_TABSTOP | cs

    dlg.append([130, staticText, 1000, (7, 7, w - 7, h - 32), cs | win32con.SS_LEFT])

    #    dlg.append([128,
    # 		"Cancel",
    # 		win32con.IDCANCEL,
    # 		(w - 60, h - 18, 50, 14), s | win32con.BS_PUSHBUTTON])

    return dlg


class CStatusProgressDialog(dialog.Dialog):
    def __init__(self, title, msg="", maxticks=100, tickincr=1):
        self.initMsg = msg
        templ = MakeProgressDlgTemplate(title, msg)
        dialog.Dialog.__init__(self, templ)
        self.maxticks = maxticks
        self.tickincr = tickincr
        self.pbar = None

    def OnInitDialog(self):
        rc = dialog.Dialog.OnInitDialog(self)
        self.static = self.GetDlgItem(1000)
        self.pbar = win32ui.CreateProgressCtrl()
        self.pbar.CreateWindow(
            win32con.WS_CHILD | win32con.WS_VISIBLE, (10, 30, 310, 44), self, 1001
        )
        self.pbar.SetRange(0, self.maxticks)
        self.pbar.SetStep(self.tickincr)
        self.progress = 0
        self.pincr = 5
        return rc

    def Close(self):
        self.EndDialog(0)

    def SetMaxTicks(self, maxticks):
        if self.pbar is not None:
            self.pbar.SetRange(0, maxticks)

    def Tick(self):
        if self.pbar is not None:
            self.pbar.StepIt()

    def SetTitle(self, text):
        self.SetWindowText(text)

    def SetText(self, text):
        self.SetDlgItemText(1000, text)

    def Set(self, pos, max=None):
        if self.pbar is not None:
            self.pbar.SetPos(pos)
            if max is not None:
                self.pbar.SetRange(0, max)


# a progress dialog created in a new thread - especially suitable for
# console apps with no message loop.
MYWM_SETTITLE = win32con.WM_USER + 10
MYWM_SETMSG = win32con.WM_USER + 11
MYWM_TICK = win32con.WM_USER + 12
MYWM_SETMAXTICKS = win32con.WM_USER + 13
MYWM_SET = win32con.WM_USER + 14


class CThreadedStatusProcessDialog(CStatusProgressDialog):
    def __init__(self, title, msg="", maxticks=100, tickincr=1):
        self.title = title
        self.msg = msg
        self.threadid = win32api.GetCurrentThreadId()
        CStatusProgressDialog.__init__(self, title, msg, maxticks, tickincr)

    def OnInitDialog(self):
        rc = CStatusProgressDialog.OnInitDialog(self)
        self.HookMessage(self.OnTitle, MYWM_SETTITLE)
        self.HookMessage(self.OnMsg, MYWM_SETMSG)
        self.HookMessage(self.OnTick, MYWM_TICK)
        self.HookMessage(self.OnMaxTicks, MYWM_SETMAXTICKS)
        self.HookMessage(self.OnSet, MYWM_SET)
        return rc

    def _Send(self, msg):
        try:
            self.PostMessage(msg)
        except win32ui.error:
            # the user closed the window - but this does not cancel the
            # process - so just ignore it.
            pass

    def OnTitle(self, msg):
        CStatusProgressDialog.SetTitle(self, self.title)

    def OnMsg(self, msg):
        CStatusProgressDialog.SetText(self, self.msg)

    def OnTick(self, msg):
        CStatusProgressDialog.Tick(self)

    def OnMaxTicks(self, msg):
        CStatusProgressDialog.SetMaxTicks(self, self.maxticks)

    def OnSet(self, msg):
        CStatusProgressDialog.Set(self, self.pos, self.max)

    def Close(self):
        assert self.threadid, "No thread!"
        win32api.PostThreadMessage(self.threadid, win32con.WM_QUIT, 0, 0)

    def SetMaxTicks(self, maxticks):
        self.maxticks = maxticks
        self._Send(MYWM_SETMAXTICKS)

    def SetTitle(self, title):
        self.title = title
        self._Send(MYWM_SETTITLE)

    def SetText(self, text):
        self.msg = text
        self._Send(MYWM_SETMSG)

    def Tick(self):
        self._Send(MYWM_TICK)

    def Set(self, pos, max=None):
        self.pos = pos
        self.max = max
        self._Send(MYWM_SET)


class ProgressThread(WinThread):
    def __init__(self, title, msg="", maxticks=100, tickincr=1):
        self.title = title
        self.msg = msg
        self.maxticks = maxticks
        self.tickincr = tickincr
        self.dialog = None
        WinThread.__init__(self)
        self.createdEvent = threading.Event()

    def InitInstance(self):
        self.dialog = CThreadedStatusProcessDialog(
            self.title, self.msg, self.maxticks, self.tickincr
        )
        self.dialog.CreateWindow()
        try:
            self.dialog.SetForegroundWindow()
        except win32ui.error:
            pass
        self.createdEvent.set()
        return WinThread.InitInstance(self)

    def ExitInstance(self):
        return 0


def StatusProgressDialog(title, msg="", maxticks=100, parent=None):
    d = CStatusProgressDialog(title, msg, maxticks)
    d.CreateWindow(parent)
    return d


def ThreadedStatusProgressDialog(title, msg="", maxticks=100):
    t = ProgressThread(title, msg, maxticks)
    t.CreateThread()
    # Need to run a basic "PumpWaitingMessages" loop just incase we are
    # running inside Pythonwin.
    # Basic timeout incase things go terribly wrong.  Ideally we should use
    # win32event.MsgWaitForMultipleObjects(), but we use a threading module
    # event - so use a dumb strategy
    end_time = time.time() + 10
    while time.time() < end_time:
        if t.createdEvent.isSet():
            break
        win32ui.PumpWaitingMessages()
        time.sleep(0.1)
    return t.dialog


def demo():
    d = StatusProgressDialog("A Demo", "Doing something...")
    import win32api

    for i in range(100):
        if i == 50:
            d.SetText("Getting there...")
        if i == 90:
            d.SetText("Nearly done...")
        win32api.Sleep(20)
        d.Tick()
    d.Close()


def thread_demo():
    d = ThreadedStatusProgressDialog("A threaded demo", "Doing something")
    import win32api

    for i in range(100):
        if i == 50:
            d.SetText("Getting there...")
        if i == 90:
            d.SetText("Nearly done...")
        win32api.Sleep(20)
        d.Tick()
    d.Close()


if __name__ == "__main__":
    thread_demo()
    # demo()



# A demo which creates a view and a frame which displays a PPM format bitmap
#
# This hasnnt been run in a while, as I dont have many of that format around!
import win32ui
import win32con
import win32api
import string


class DIBView:
    def __init__(self, doc, dib):
        self.dib = dib
        self.view = win32ui.CreateView(doc)
        self.width = self.height = 0
        # set up message handlers
        # 		self.view.OnPrepareDC = self.OnPrepareDC
        self.view.HookMessage(self.OnSize, win32con.WM_SIZE)

    def OnSize(self, params):
        lParam = params[3]
        self.width = win32api.LOWORD(lParam)
        self.height = win32api.HIWORD(lParam)

    def OnDraw(self, ob, dc):
        # set sizes used for "non strecth" mode.
        self.view.SetScrollSizes(win32con.MM_TEXT, self.dib.GetSize())
        dibSize = self.dib.GetSize()
        dibRect = (0, 0, dibSize[0], dibSize[1])
        # stretch BMP.
        # self.dib.Paint(dc, (0,0,self.width, self.height),dibRect)
        # non stretch.
        self.dib.Paint(dc)


class DIBDemo:
    def __init__(self, filename, *bPBM):
        # init data members
        f = open(filename, "rb")
        dib = win32ui.CreateDIBitmap()
        if len(bPBM) > 0:
            magic = f.readline()
            if magic != "P6\n":
                print("The file is not a PBM format file")
                raise ValueError("Failed - The file is not a PBM format file")
            # check magic?
            rowcollist = f.readline().split()
            cols = int(rowcollist[0])
            rows = int(rowcollist[1])
            f.readline()  # whats this one?
            dib.LoadPBMData(f, (cols, rows))
        else:
            dib.LoadWindowsFormatFile(f)
        f.close()
        # create doc/view
        self.doc = win32ui.CreateDoc()
        self.dibView = DIBView(self.doc, dib)
        self.frame = win32ui.CreateMDIFrame()
        self.frame.LoadFrame()  # this will force OnCreateClient
        self.doc.SetTitle("DIB Demo")
        self.frame.ShowWindow()

        # display the sucka
        self.frame.ActivateFrame()

    def OnCreateClient(self, createparams, context):
        self.dibView.view.CreateWindow(self.frame)
        return 1


if __name__ == "__main__":
    import demoutils

    demoutils.NotAScript()



# Contributed by Kelly Kranabetter.
import os, sys
import win32security, ntsecuritycon, pywintypes, winerror

# get security information
# name=r"c:\autoexec.bat"
# name= r"g:\!workgrp\lim"
name = sys.argv[0]

if not os.path.exists(name):
    print(name, "does not exist!")
    sys.exit()

print("On file ", name, "\n")

# get owner SID
print("OWNER")
try:
    sd = win32security.GetFileSecurity(name, win32security.OWNER_SECURITY_INFORMATION)
    sid = sd.GetSecurityDescriptorOwner()
    print("  ", win32security.LookupAccountSid(None, sid))
except pywintypes.error as exc:
    # in automation and network shares we see:
    # pywintypes.error: (1332, 'LookupAccountName', 'No mapping between account names and security IDs was done.')
    if exc.winerror != winerror.ERROR_NONE_MAPPED:
        raise
    print("No owner information is available")

# get group SID
try:
    print("GROUP")
    sd = win32security.GetFileSecurity(name, win32security.GROUP_SECURITY_INFORMATION)
    sid = sd.GetSecurityDescriptorGroup()
    print("  ", win32security.LookupAccountSid(None, sid))
except pywintypes.error as exc:
    if exc.winerror != winerror.ERROR_NONE_MAPPED:
        raise
    print("No group information is available")

# get ACEs
sd = win32security.GetFileSecurity(name, win32security.DACL_SECURITY_INFORMATION)
dacl = sd.GetSecurityDescriptorDacl()
if dacl == None:
    print("No Discretionary ACL")
else:
    for ace_no in range(0, dacl.GetAceCount()):
        ace = dacl.GetAce(ace_no)
        print("ACE", ace_no)

        print("  -Type")
        for i in (
            "ACCESS_ALLOWED_ACE_TYPE",
            "ACCESS_DENIED_ACE_TYPE",
            "SYSTEM_AUDIT_ACE_TYPE",
            "SYSTEM_ALARM_ACE_TYPE",
        ):
            if getattr(ntsecuritycon, i) == ace[0][0]:
                print("    ", i)

        print("  -Flags", hex(ace[0][1]))
        for i in (
            "OBJECT_INHERIT_ACE",
            "CONTAINER_INHERIT_ACE",
            "NO_PROPAGATE_INHERIT_ACE",
            "INHERIT_ONLY_ACE",
            "SUCCESSFUL_ACCESS_ACE_FLAG",
            "FAILED_ACCESS_ACE_FLAG",
        ):
            if getattr(ntsecuritycon, i) & ace[0][1] == getattr(ntsecuritycon, i):
                print("    ", i)

        print("  -mask", hex(ace[1]))

        # files and directories do permissions differently
        permissions_file = (
            "DELETE",
            "READ_CONTROL",
            "WRITE_DAC",
            "WRITE_OWNER",
            "SYNCHRONIZE",
            "FILE_GENERIC_READ",
            "FILE_GENERIC_WRITE",
            "FILE_GENERIC_EXECUTE",
            "FILE_DELETE_CHILD",
        )
        permissions_dir = (
            "DELETE",
            "READ_CONTROL",
            "WRITE_DAC",
            "WRITE_OWNER",
            "SYNCHRONIZE",
            "FILE_ADD_SUBDIRECTORY",
            "FILE_ADD_FILE",
            "FILE_DELETE_CHILD",
            "FILE_LIST_DIRECTORY",
            "FILE_TRAVERSE",
            "FILE_READ_ATTRIBUTES",
            "FILE_WRITE_ATTRIBUTES",
            "FILE_READ_EA",
            "FILE_WRITE_EA",
        )
        permissions_dir_inherit = (
            "DELETE",
            "READ_CONTROL",
            "WRITE_DAC",
            "WRITE_OWNER",
            "SYNCHRONIZE",
            "GENERIC_READ",
            "GENERIC_WRITE",
            "GENERIC_EXECUTE",
            "GENERIC_ALL",
        )
        if os.path.isfile(name):
            permissions = permissions_file
        else:
            permissions = permissions_dir
            # directories also contain an ACE that is inherited by children (files) within them
            if (
                ace[0][1] & ntsecuritycon.OBJECT_INHERIT_ACE
                == ntsecuritycon.OBJECT_INHERIT_ACE
                and ace[0][1] & ntsecuritycon.INHERIT_ONLY_ACE
                == ntsecuritycon.INHERIT_ONLY_ACE
            ):
                permissions = permissions_dir_inherit

        calc_mask = 0  # calculate the mask so we can see if we are printing all of the permissions
        for i in permissions:
            if getattr(ntsecuritycon, i) & ace[1] == getattr(ntsecuritycon, i):
                calc_mask = calc_mask | getattr(ntsecuritycon, i)
                print("    ", i)
        print("  ", "Calculated Check Mask=", hex(calc_mask))
        print("  -SID\n    ", win32security.LookupAccountSid(None, ace[2]))



# cmdline - command line utilities.
import sys
import win32ui
import string


def ParseArgs(str):
    import string

    ret = []
    pos = 0
    length = len(str)
    while pos < length:
        try:
            while str[pos] in string.whitespace:
                pos = pos + 1
        except IndexError:
            break
        if pos >= length:
            break
        if str[pos] == '"':
            pos = pos + 1
            try:
                endPos = str.index('"', pos) - 1
                nextPos = endPos + 2
            except ValueError:
                endPos = length
                nextPos = endPos + 1
        else:
            endPos = pos
            while endPos < length and not str[endPos] in string.whitespace:
                endPos = endPos + 1
            nextPos = endPos + 1
        ret.append(str[pos : endPos + 1].strip())
        pos = nextPos
    return ret


def FixArgFileName(fileName):
    """Convert a filename on the commandline to something useful.
    Given an automatic filename on the commandline, turn it a python module name,
    with the path added to sys.path."""
    import os

    path, fname = os.path.split(fileName)
    if len(path) == 0:
        path = os.curdir
    path = os.path.abspath(path)
    # must check that the command line arg's path is in sys.path
    for syspath in sys.path:
        if os.path.abspath(syspath) == path:
            break
    else:
        sys.path.append(path)
    return os.path.splitext(fname)[0]



# -*- Mode: Python; tab-width: 4 -*-
#

# This module, and the timer.pyd core timer support, were written by
# Sam Rushing (rushing@nightmare.com)

import timer
import time

# Timers are based on Windows messages.  So we need
# to do the event-loop thing!
import win32event, win32gui

# glork holds a simple counter for us.


class glork:
    def __init__(self, delay=1000, max=10):
        self.x = 0
        self.max = max
        self.id = timer.set_timer(delay, self.increment)
        # Could use the threading module, but this is
        # a win32 extension test after all! :-)
        self.event = win32event.CreateEvent(None, 0, 0, None)

    def increment(self, id, time):
        print("x = %d" % self.x)
        self.x = self.x + 1
        # if we've reached the max count,
        # kill off the timer.
        if self.x > self.max:
            # we could have used 'self.id' here, too
            timer.kill_timer(id)
            win32event.SetEvent(self.event)


# create a counter that will count from '1' thru '10', incrementing
# once a second, and then stop.


def demo(delay=1000, stop=10):
    g = glork(delay, stop)
    # Timers are message based - so we need
    # To run a message loop while waiting for our timers
    # to expire.
    start_time = time.time()
    while 1:
        # We can't simply give a timeout of 30 seconds, as
        # we may continouusly be recieving other input messages,
        # and therefore never expire.
        rc = win32event.MsgWaitForMultipleObjects(
            (g.event,),  # list of objects
            0,  # wait all
            500,  # timeout
            win32event.QS_ALLEVENTS,  # type of input
        )
        if rc == win32event.WAIT_OBJECT_0:
            # Event signalled.
            break
        elif rc == win32event.WAIT_OBJECT_0 + 1:
            # Message waiting.
            if win32gui.PumpWaitingMessages():
                raise RuntimeError("We got an unexpected WM_QUIT message!")
        else:
            # This wait timed-out.
            if time.time() - start_time > 30:
                raise RuntimeError("We timed out waiting for the timers to expire!")


if __name__ == "__main__":
    demo()



import win32security, ntsecuritycon, winnt


class Enum:
    def __init__(self, *const_names):
        """Accepts variable number of constant names that can be found in either
        win32security, ntsecuritycon, or winnt."""
        for const_name in const_names:
            try:
                const_val = getattr(win32security, const_name)
            except AttributeError:
                try:
                    const_val = getattr(ntsecuritycon, const_name)
                except AttributeError:
                    try:
                        const_val = getattr(winnt, const_name)
                    except AttributeError:
                        raise AttributeError(
                            'Constant "%s" not found in win32security, ntsecuritycon, or winnt.'
                            % const_name
                        )
            setattr(self, const_name, const_val)

    def lookup_name(self, const_val):
        """Looks up the name of a particular value."""
        for k, v in self.__dict__.items():
            if v == const_val:
                return k
        raise AttributeError("Value %s not found in enum" % const_val)

    def lookup_flags(self, flags):
        """Returns the names of all recognized flags in input, and any flags not found in the enum."""
        flag_names = []
        unknown_flags = flags
        for k, v in self.__dict__.items():
            if flags & v == v:
                flag_names.append(k)
                unknown_flags = unknown_flags & ~v
        return flag_names, unknown_flags


TOKEN_INFORMATION_CLASS = Enum(
    "TokenUser",
    "TokenGroups",
    "TokenPrivileges",
    "TokenOwner",
    "TokenPrimaryGroup",
    "TokenDefaultDacl",
    "TokenSource",
    "TokenType",
    "TokenImpersonationLevel",
    "TokenStatistics",
    "TokenRestrictedSids",
    "TokenSessionId",
    "TokenGroupsAndPrivileges",
    "TokenSessionReference",
    "TokenSandBoxInert",
    "TokenAuditPolicy",
    "TokenOrigin",
    "TokenElevationType",
    "TokenLinkedToken",
    "TokenElevation",
    "TokenHasRestrictions",
    "TokenAccessInformation",
    "TokenVirtualizationAllowed",
    "TokenVirtualizationEnabled",
    "TokenIntegrityLevel",
    "TokenUIAccess",
    "TokenMandatoryPolicy",
    "TokenLogonSid",
)

TOKEN_TYPE = Enum("TokenPrimary", "TokenImpersonation")

TOKEN_ELEVATION_TYPE = Enum(
    "TokenElevationTypeDefault", "TokenElevationTypeFull", "TokenElevationTypeLimited"
)

POLICY_AUDIT_EVENT_TYPE = Enum(
    "AuditCategorySystem",
    "AuditCategoryLogon",
    "AuditCategoryObjectAccess",
    "AuditCategoryPrivilegeUse",
    "AuditCategoryDetailedTracking",
    "AuditCategoryPolicyChange",
    "AuditCategoryAccountManagement",
    "AuditCategoryDirectoryServiceAccess",
    "AuditCategoryAccountLogon",
)

POLICY_INFORMATION_CLASS = Enum(
    "PolicyAuditLogInformation",
    "PolicyAuditEventsInformation",
    "PolicyPrimaryDomainInformation",
    "PolicyPdAccountInformation",
    "PolicyAccountDomainInformation",
    "PolicyLsaServerRoleInformation",
    "PolicyReplicaSourceInformation",
    "PolicyDefaultQuotaInformation",
    "PolicyModificationInformation",
    "PolicyAuditFullSetInformation",
    "PolicyAuditFullQueryInformation",
    "PolicyDnsDomainInformation",
)

POLICY_LSA_SERVER_ROLE = Enum("PolicyServerRoleBackup", "PolicyServerRolePrimary")

## access modes for opening a policy handle - this is not a real enum
POLICY_ACCESS_MODES = Enum(
    "POLICY_VIEW_LOCAL_INFORMATION",
    "POLICY_VIEW_AUDIT_INFORMATION",
    "POLICY_GET_PRIVATE_INFORMATION",
    "POLICY_TRUST_ADMIN",
    "POLICY_CREATE_ACCOUNT",
    "POLICY_CREATE_SECRET",
    "POLICY_CREATE_PRIVILEGE",
    "POLICY_SET_DEFAULT_QUOTA_LIMITS",
    "POLICY_SET_AUDIT_REQUIREMENTS",
    "POLICY_AUDIT_LOG_ADMIN",
    "POLICY_SERVER_ADMIN",
    "POLICY_LOOKUP_NAMES",
    "POLICY_NOTIFICATION",
    "POLICY_ALL_ACCESS",
    "POLICY_READ",
    "POLICY_WRITE",
    "POLICY_EXECUTE",
)

## EventAuditingOptions flags - not a real enum
POLICY_AUDIT_EVENT_OPTIONS_FLAGS = Enum(
    "POLICY_AUDIT_EVENT_UNCHANGED",
    "POLICY_AUDIT_EVENT_SUCCESS",
    "POLICY_AUDIT_EVENT_FAILURE",
    "POLICY_AUDIT_EVENT_NONE",
)

# AceType in ACE_HEADER - not a real enum
ACE_TYPE = Enum(
    "ACCESS_MIN_MS_ACE_TYPE",
    "ACCESS_ALLOWED_ACE_TYPE",
    "ACCESS_DENIED_ACE_TYPE",
    "SYSTEM_AUDIT_ACE_TYPE",
    "SYSTEM_ALARM_ACE_TYPE",
    "ACCESS_MAX_MS_V2_ACE_TYPE",
    "ACCESS_ALLOWED_COMPOUND_ACE_TYPE",
    "ACCESS_MAX_MS_V3_ACE_TYPE",
    "ACCESS_MIN_MS_OBJECT_ACE_TYPE",
    "ACCESS_ALLOWED_OBJECT_ACE_TYPE",
    "ACCESS_DENIED_OBJECT_ACE_TYPE",
    "SYSTEM_AUDIT_OBJECT_ACE_TYPE",
    "SYSTEM_ALARM_OBJECT_ACE_TYPE",
    "ACCESS_MAX_MS_OBJECT_ACE_TYPE",
    "ACCESS_MAX_MS_V4_ACE_TYPE",
    "ACCESS_MAX_MS_ACE_TYPE",
    "ACCESS_ALLOWED_CALLBACK_ACE_TYPE",
    "ACCESS_DENIED_CALLBACK_ACE_TYPE",
    "ACCESS_ALLOWED_CALLBACK_OBJECT_ACE_TYPE",
    "ACCESS_DENIED_CALLBACK_OBJECT_ACE_TYPE",
    "SYSTEM_AUDIT_CALLBACK_ACE_TYPE",
    "SYSTEM_ALARM_CALLBACK_ACE_TYPE",
    "SYSTEM_AUDIT_CALLBACK_OBJECT_ACE_TYPE",
    "SYSTEM_ALARM_CALLBACK_OBJECT_ACE_TYPE",
    "SYSTEM_MANDATORY_LABEL_ACE_TYPE",
    "ACCESS_MAX_MS_V5_ACE_TYPE",
)

# bit flags for AceFlags - not a real enum
ACE_FLAGS = Enum(
    "CONTAINER_INHERIT_ACE",
    "FAILED_ACCESS_ACE_FLAG",
    "INHERIT_ONLY_ACE",
    "INHERITED_ACE",
    "NO_PROPAGATE_INHERIT_ACE",
    "OBJECT_INHERIT_ACE",
    "SUCCESSFUL_ACCESS_ACE_FLAG",
    "NO_INHERITANCE",
    "SUB_CONTAINERS_AND_OBJECTS_INHERIT",
    "SUB_CONTAINERS_ONLY_INHERIT",
    "SUB_OBJECTS_ONLY_INHERIT",
)

# used in SetEntriesInAcl - very similar to ACE_TYPE
ACCESS_MODE = Enum(
    "NOT_USED_ACCESS",
    "GRANT_ACCESS",
    "SET_ACCESS",
    "DENY_ACCESS",
    "REVOKE_ACCESS",
    "SET_AUDIT_SUCCESS",
    "SET_AUDIT_FAILURE",
)

# Bit flags in PSECURITY_DESCRIPTOR->Control - not a real enum
SECURITY_DESCRIPTOR_CONTROL_FLAGS = Enum(
    "SE_DACL_AUTO_INHERITED",  ## win2k and up
    "SE_SACL_AUTO_INHERITED",  ## win2k and up
    "SE_DACL_PROTECTED",  ## win2k and up
    "SE_SACL_PROTECTED",  ## win2k and up
    "SE_DACL_DEFAULTED",
    "SE_DACL_PRESENT",
    "SE_GROUP_DEFAULTED",
    "SE_OWNER_DEFAULTED",
    "SE_SACL_PRESENT",
    "SE_SELF_RELATIVE",
    "SE_SACL_DEFAULTED",
)

# types of SID
SID_NAME_USE = Enum(
    "SidTypeUser",
    "SidTypeGroup",
    "SidTypeDomain",
    "SidTypeAlias",
    "SidTypeWellKnownGroup",
    "SidTypeDeletedAccount",
    "SidTypeInvalid",
    "SidTypeUnknown",
    "SidTypeComputer",
    "SidTypeLabel",
)

## bit flags, not a real enum
TOKEN_ACCESS_PRIVILEGES = Enum(
    "TOKEN_ADJUST_DEFAULT",
    "TOKEN_ADJUST_GROUPS",
    "TOKEN_ADJUST_PRIVILEGES",
    "TOKEN_ALL_ACCESS",
    "TOKEN_ASSIGN_PRIMARY",
    "TOKEN_DUPLICATE",
    "TOKEN_EXECUTE",
    "TOKEN_IMPERSONATE",
    "TOKEN_QUERY",
    "TOKEN_QUERY_SOURCE",
    "TOKEN_READ",
    "TOKEN_WRITE",
)

SECURITY_IMPERSONATION_LEVEL = Enum(
    "SecurityAnonymous",
    "SecurityIdentification",
    "SecurityImpersonation",
    "SecurityDelegation",
)

POLICY_SERVER_ENABLE_STATE = Enum("PolicyServerEnabled", "PolicyServerDisabled")

POLICY_NOTIFICATION_INFORMATION_CLASS = Enum(
    "PolicyNotifyAuditEventsInformation",
    "PolicyNotifyAccountDomainInformation",
    "PolicyNotifyServerRoleInformation",
    "PolicyNotifyDnsDomainInformation",
    "PolicyNotifyDomainEfsInformation",
    "PolicyNotifyDomainKerberosTicketInformation",
    "PolicyNotifyMachineAccountPasswordInformation",
)

TRUSTED_INFORMATION_CLASS = Enum(
    "TrustedDomainNameInformation",
    "TrustedControllersInformation",
    "TrustedPosixOffsetInformation",
    "TrustedPasswordInformation",
    "TrustedDomainInformationBasic",
    "TrustedDomainInformationEx",
    "TrustedDomainAuthInformation",
    "TrustedDomainFullInformation",
    "TrustedDomainAuthInformationInternal",
    "TrustedDomainFullInformationInternal",
    "TrustedDomainInformationEx2Internal",
    "TrustedDomainFullInformation2Internal",
)

TRUSTEE_FORM = Enum(
    "TRUSTEE_IS_SID",
    "TRUSTEE_IS_NAME",
    "TRUSTEE_BAD_FORM",
    "TRUSTEE_IS_OBJECTS_AND_SID",
    "TRUSTEE_IS_OBJECTS_AND_NAME",
)

TRUSTEE_TYPE = Enum(
    "TRUSTEE_IS_UNKNOWN",
    "TRUSTEE_IS_USER",
    "TRUSTEE_IS_GROUP",
    "TRUSTEE_IS_DOMAIN",
    "TRUSTEE_IS_ALIAS",
    "TRUSTEE_IS_WELL_KNOWN_GROUP",
    "TRUSTEE_IS_DELETED",
    "TRUSTEE_IS_INVALID",
    "TRUSTEE_IS_COMPUTER",
)

## SE_OBJECT_TYPE - securable objects
SE_OBJECT_TYPE = Enum(
    "SE_UNKNOWN_OBJECT_TYPE",
    "SE_FILE_OBJECT",
    "SE_SERVICE",
    "SE_PRINTER",
    "SE_REGISTRY_KEY",
    "SE_LMSHARE",
    "SE_KERNEL_OBJECT",
    "SE_WINDOW_OBJECT",
    "SE_DS_OBJECT",
    "SE_DS_OBJECT_ALL",
    "SE_PROVIDER_DEFINED_OBJECT",
    "SE_WMIGUID_OBJECT",
    "SE_REGISTRY_WOW64_32KEY",
)

PRIVILEGE_FLAGS = Enum(
    "SE_PRIVILEGE_ENABLED_BY_DEFAULT",
    "SE_PRIVILEGE_ENABLED",
    "SE_PRIVILEGE_USED_FOR_ACCESS",
)

# Group flags used with TokenGroups
TOKEN_GROUP_ATTRIBUTES = Enum(
    "SE_GROUP_MANDATORY",
    "SE_GROUP_ENABLED_BY_DEFAULT",
    "SE_GROUP_ENABLED",
    "SE_GROUP_OWNER",
    "SE_GROUP_USE_FOR_DENY_ONLY",
    "SE_GROUP_INTEGRITY",
    "SE_GROUP_INTEGRITY_ENABLED",
    "SE_GROUP_LOGON_ID",
    "SE_GROUP_RESOURCE",
)

# Privilege flags returned by TokenPrivileges
TOKEN_PRIVILEGE_ATTRIBUTES = Enum(
    "SE_PRIVILEGE_ENABLED_BY_DEFAULT",
    "SE_PRIVILEGE_ENABLED",
    "SE_PRIVILEGE_REMOVED",
    "SE_PRIVILEGE_USED_FOR_ACCESS",
)



from datetime import datetime
import unittest

import pywintypes
import win32com.client
import win32com.test.util
import win32com.server.util
from win32timezone import TimeZoneInfo

# A COM object so we can pass dates to and from the COM boundary.
class Tester:
    _public_methods_ = ["TestDate"]

    def TestDate(self, d):
        assert isinstance(d, datetime)
        return d


def test_ob():
    return win32com.client.Dispatch(win32com.server.util.wrap(Tester()))


class TestCase(win32com.test.util.TestCase):
    def check(self, d, expected=None):
        if not issubclass(pywintypes.TimeType, datetime):
            self.skipTest("this is testing pywintypes and datetime")
        got = test_ob().TestDate(d)
        self.assertEqual(got, expected or d)

    def testUTC(self):
        self.check(
            datetime(
                year=2000,
                month=12,
                day=25,
                microsecond=500000,
                tzinfo=TimeZoneInfo.utc(),
            )
        )

    def testLocal(self):
        self.check(
            datetime(
                year=2000,
                month=12,
                day=25,
                microsecond=500000,
                tzinfo=TimeZoneInfo.local(),
            )
        )

    def testMSTruncated(self):
        # milliseconds are kept but microseconds are lost after rounding.
        self.check(
            datetime(
                year=2000,
                month=12,
                day=25,
                microsecond=500500,
                tzinfo=TimeZoneInfo.utc(),
            ),
            datetime(
                year=2000,
                month=12,
                day=25,
                microsecond=500000,
                tzinfo=TimeZoneInfo.utc(),
            ),
        )


if __name__ == "__main__":
    unittest.main()



## Demonstrates a "push" subscription with a callback function
import win32evtlog

query_text = '*[System[Provider[@Name="Microsoft-Windows-Winlogon"]]]'


def c(reason, context, evt):
    if reason == win32evtlog.EvtSubscribeActionError:
        print("EvtSubscribeActionError")
    elif reason == win32evtlog.EvtSubscribeActionDeliver:
        print("EvtSubscribeActionDeliver")
    else:
        print("??? Unknown action ???", reason)
    context.append(win32evtlog.EvtRender(evt, win32evtlog.EvtRenderEventXml))
    return 0


evttext = []
s = win32evtlog.EvtSubscribe(
    "System",
    win32evtlog.EvtSubscribeStartAtOldestRecord,
    Query="*",
    Callback=c,
    Context=evttext,
)



# The start of a win32gui generic demo.
# Feel free to contribute more demos back ;-)

import win32gui, win32con, win32api
import time, math, random


def _MyCallback(hwnd, extra):
    hwnds, classes = extra
    hwnds.append(hwnd)
    classes[win32gui.GetClassName(hwnd)] = 1


def TestEnumWindows():
    windows = []
    classes = {}
    win32gui.EnumWindows(_MyCallback, (windows, classes))
    print(
        "Enumerated a total of %d windows with %d classes"
        % (len(windows), len(classes))
    )
    if "tooltips_class32" not in classes:
        print("Hrmmmm - I'm very surprised to not find a 'tooltips_class32' class.")


def OnPaint_1(hwnd, msg, wp, lp):
    dc, ps = win32gui.BeginPaint(hwnd)
    win32gui.SetGraphicsMode(dc, win32con.GM_ADVANCED)
    br = win32gui.CreateSolidBrush(win32api.RGB(255, 0, 0))
    win32gui.SelectObject(dc, br)
    angle = win32gui.GetWindowLong(hwnd, win32con.GWL_USERDATA)
    win32gui.SetWindowLong(hwnd, win32con.GWL_USERDATA, angle + 2)
    r_angle = angle * (math.pi / 180)
    win32gui.SetWorldTransform(
        dc,
        {
            "M11": math.cos(r_angle),
            "M12": math.sin(r_angle),
            "M21": math.sin(r_angle) * -1,
            "M22": math.cos(r_angle),
            "Dx": 250,
            "Dy": 250,
        },
    )
    win32gui.MoveToEx(dc, 250, 250)
    win32gui.BeginPath(dc)
    win32gui.Pie(dc, 10, 70, 200, 200, 350, 350, 75, 10)
    win32gui.Chord(dc, 200, 200, 850, 0, 350, 350, 75, 10)
    win32gui.LineTo(dc, 300, 300)
    win32gui.LineTo(dc, 100, 20)
    win32gui.LineTo(dc, 20, 100)
    win32gui.LineTo(dc, 400, 0)
    win32gui.LineTo(dc, 0, 400)
    win32gui.EndPath(dc)
    win32gui.StrokeAndFillPath(dc)
    win32gui.EndPaint(hwnd, ps)
    return 0


wndproc_1 = {win32con.WM_PAINT: OnPaint_1}


def OnPaint_2(hwnd, msg, wp, lp):
    dc, ps = win32gui.BeginPaint(hwnd)
    win32gui.SetGraphicsMode(dc, win32con.GM_ADVANCED)
    l, t, r, b = win32gui.GetClientRect(hwnd)

    for x in range(25):
        vertices = (
            {
                "x": int(random.random() * r),
                "y": int(random.random() * b),
                "Red": int(random.random() * 0xFF00),
                "Green": 0,
                "Blue": 0,
                "Alpha": 0,
            },
            {
                "x": int(random.random() * r),
                "y": int(random.random() * b),
                "Red": 0,
                "Green": int(random.random() * 0xFF00),
                "Blue": 0,
                "Alpha": 0,
            },
            {
                "x": int(random.random() * r),
                "y": int(random.random() * b),
                "Red": 0,
                "Green": 0,
                "Blue": int(random.random() * 0xFF00),
                "Alpha": 0,
            },
        )
        mesh = ((0, 1, 2),)
        win32gui.GradientFill(dc, vertices, mesh, win32con.GRADIENT_FILL_TRIANGLE)
    win32gui.EndPaint(hwnd, ps)
    return 0


wndproc_2 = {win32con.WM_PAINT: OnPaint_2}


def TestSetWorldTransform():
    wc = win32gui.WNDCLASS()
    wc.lpszClassName = "test_win32gui_1"
    wc.style = win32con.CS_GLOBALCLASS | win32con.CS_VREDRAW | win32con.CS_HREDRAW
    wc.hbrBackground = win32con.COLOR_WINDOW + 1
    wc.lpfnWndProc = wndproc_1
    class_atom = win32gui.RegisterClass(wc)
    hwnd = win32gui.CreateWindow(
        wc.lpszClassName,
        "Spin the Lobster!",
        win32con.WS_CAPTION | win32con.WS_VISIBLE,
        100,
        100,
        900,
        900,
        0,
        0,
        0,
        None,
    )
    for x in range(500):
        win32gui.InvalidateRect(hwnd, None, True)
        win32gui.PumpWaitingMessages()
        time.sleep(0.01)
    win32gui.DestroyWindow(hwnd)
    win32gui.UnregisterClass(wc.lpszClassName, None)


def TestGradientFill():
    wc = win32gui.WNDCLASS()
    wc.lpszClassName = "test_win32gui_2"
    wc.style = win32con.CS_GLOBALCLASS | win32con.CS_VREDRAW | win32con.CS_HREDRAW
    wc.hbrBackground = win32con.COLOR_WINDOW + 1
    wc.lpfnWndProc = wndproc_2
    class_atom = win32gui.RegisterClass(wc)
    hwnd = win32gui.CreateWindowEx(
        0,
        class_atom,
        "Kaleidoscope",
        win32con.WS_CAPTION
        | win32con.WS_VISIBLE
        | win32con.WS_THICKFRAME
        | win32con.WS_SYSMENU,
        100,
        100,
        900,
        900,
        0,
        0,
        0,
        None,
    )
    s = win32gui.GetWindowLong(hwnd, win32con.GWL_EXSTYLE)
    win32gui.SetWindowLong(hwnd, win32con.GWL_EXSTYLE, s | win32con.WS_EX_LAYERED)
    win32gui.SetLayeredWindowAttributes(hwnd, 0, 175, win32con.LWA_ALPHA)
    for x in range(30):
        win32gui.InvalidateRect(hwnd, None, True)
        win32gui.PumpWaitingMessages()
        time.sleep(0.3)
    win32gui.DestroyWindow(hwnd)
    win32gui.UnregisterClass(class_atom, None)


print("Enumerating all windows...")
TestEnumWindows()
print("Testing drawing functions ...")
TestSetWorldTransform()
TestGradientFill()
print("All tests done!")
